#
# This file is part of LiteX.
#
# Copyright (c) 2020 bunnie <bunnie@kosagi.com>
# SPDX-License-Identifier: BSD-2-Clause

from migen.genlib.cdc import MultiReg
from migen.genlib.fifo import SyncFIFOBuffered

from litex.soc.interconnect import wishbone
from litex.soc.interconnect.csr_eventmanager import *

from litex.soc.integration.doc import AutoDoc, ModuleDoc


class S7SPIOPI(Module, AutoCSR, AutoDoc):
    def __init__(self, pads,
        dq_delay_taps  = 0,
        sclk_name      = "SCLK_ODDR",
        iddr_name      = "SPI_IDDR",
        cipo_name      = "CIPO_FDRE",
        sim            = False,
        spiread        = False,
        prefetch_lines = 1):

        self.sclk_name = sclk_name
        self.iddr_name = iddr_name
        self.cipo_name = cipo_name
        self.spiread = spiread
        self.gsr = Signal()
        self.keyclearb = Signal(reset=1)
        self.cfgmclk = Signal()

        self.dq = dq = TSTriple(7) # dq[0] is special because it is also copi
        self.dq_copi = dq_copi = TSTriple(1) # this has similar structure but an independent "oe" signal

        self.intro = ModuleDoc("""Intro

        SpiOpi implements a dual-mode SPI or OPI interface. OPI is an octal (8-bit) wide variant of
        SPI, which is unique to Macronix parts. It is concurrently interoperable with SPI. The chip
        supports "DTR mode" (double transfer rate, e.g. DDR) where data is transferred on each edge
        of the clock, and there is a source-synchronous DQS associated with the input data.

        The chip by default boots into SPI-only mode (unless NV bits are burned otherwise) so to
        enable OPI, a config register needs to be written with SPI mode. Note that once the config
        register is written, the only way to return to SPI mode is to change it with OPI writes, or
        to issue a hardware reset. This has major implications for reconfiguring the FPGA: a simple
        JTAG command to reload from SPI will not yank PROG_B low, and so the SPI ROM will be in DOPI,
        and SPI loading will fail. Thus, system architects must take into consideration a hard reset
        for the ROM whenever a bitstream reload is demanded of the FPGA.

        The SpiOpi architecture is split into two levels: a command manager, and a cycle manager. The
        command manager is responsible for taking the current wishbone request and CSR state and
        unpacking these into cycle-by-cycle requests. The cycle manager is responsible for coordinating
        the cycle-by-cycle requests.

        In SPI mode, this means marshalling byte-wide requests into a series of 8 serial cyles.

        In OPI [DOPI] mode, this means marshalling 16-bit wide requests into a pair of back-to-back
        DDR cycles. Note that because the cycles are DDR, this means one 16-bit wide request must be
        issued every cycle to keep up with the interface.

        For the output of data to ROM, expects a clock called "spinor_delayed" which is a delayed
        version of "sys". The delay is necessary to get the correct phase relationship between the
        SIO and SCLK in DTR/DDR mode, and it also has to compensate for the special-case difference
        in the CCLK pad vs other I/O.

        For the input, DQS signal is independently delayed relative to the DQ signals using an IDELAYE2
        block. At a REFCLK frequency of 200 MHz, each delay tap adds 78ps, so up to a 2.418ns delay is
        possible between DQS and DQ. The goal is to delay DQS relative to DQ, because the SPI chip
        launches both with concurrent rising edges (to within 0.6ns), but the IDDR register needs the
        rising edge of DQS to be centered inside the DQ eye.

        In DOPI mode, there is a prefetch buffer. It will read `prefetch_lines` cache lines of data
        into the prefetch buffer. A cache line is 256 bits (or 8x32-bit words). The maximum value is
        63 lines (one line is necessary for synchronization margin). The downside of setting
        prefetch_lines high is that the prefetcher is running constantly and burning power, while
        throwing away most data. In practice, the CPU will typically consume data at only slightly
        faster than the rate of read-out from DOPI-mode ROM, and once data is consumed the prefetch
        resumes. Thus, prefetch_lines is probably optimally around 1-3 lines read-ahead of the CPU.
        Any higher than 3 lines probably just wastes power. In short simulations, 1 line of prefetch
        seems to be enough to keep the prefetcher ahead of the CPU even when it's simply running
        straight-line code.

        Note the "sim" parameter exists because there seems to be a bug in xvlog that doesn't
        correctly simulate the IDELAY machines. Setting "sim" to True removes the IDELAY machines
        and passes the data through directly, but in real hardware the IDELAY machines are necessary
        to meet timing between DQS and DQ.

        dq_delay_taps probably doesn't need to be adjusted; it can be tweaked for timing closure. The
        delays can also be adjusted at runtime.
        """)

        if sim == False:
            idelay_name = "IDELAYE2"
            bufr_name = "BUFG" # we actually want a slightly slower buffer here...
        else:
            idelay_name = "IDELAYE2_SIM"
            bufr_name = "BUFR_SIM"

        if prefetch_lines > 62:
            prefetch_lines = 62

        self.spi_mode = spi_mode = Signal(reset=1) # When reset is asserted, force into spi mode
        cs_n = Signal(reset=1) # Make sure CS is sane on reset, too

        self.config = CSRStorage(fields=[
            CSRField("dummy", size=5, description="Number of dummy cycles", reset=10),
        ])

        delay_type="VAR_LOAD"

        # DQS input conditioning -----------------------------------------------------------------
        dqs_iobuf = Signal()
        self.clock_domains.cd_dqs = ClockDomain(reset_less=True)
        self.comb += self.cd_dqs.clk.eq(dqs_iobuf)
        self.specials += [
            Instance(bufr_name, i_I=pads.dqs, o_O=dqs_iobuf),
        ]

        # DQ connections -------------------------------------------------------------------------
        # PHY API
        self.do = Signal(16) # OPI data to SPI
        self.di = Signal(16) # OPI data from SPI
        self.tx = Signal()   # When asserted OPI is transmitting data to SPI, otherwise, receiving

        self.copi = Signal() # SPI data to SPI
        self.cipo = Signal() # SPI data from SPI

        # Delay programming API
        self.delay_config = CSRStorage(fields=[
            CSRField("d",    size=5, description="Delay amount; each increment is 78ps", reset=dq_delay_taps),
            CSRField("load", size=1, description="Force delay taps to delay_d"),
        ])
        self.delay_status = CSRStatus(fields=[
            CSRField("q", size=5, description="Readback of current delay amount, useful if inc/ce is used to set"),
        ])
        self.delay_update  = Signal()
        self.hw_delay_load = Signal(reset=1) # latch in the initial value on reset
        reset_counter = Signal(4, reset=15)
        self.sync += \
            If(reset_counter != 0,
                reset_counter.eq(reset_counter - 1)
            ).Else(
                self.hw_delay_load.eq(0)
            )
        self.sync += self.delay_update.eq(self.hw_delay_load | self.delay_config.fields.load)

        # Break system API into rising/falling edge samples
        do_rise = Signal(8) # data output presented on the rising edge
        do_fall = Signal(8) # data output presented on the falling edge
        self.comb += [do_rise.eq(self.do[8:]), do_fall.eq(self.do[:8])]

        di_rise = Signal(8)
        di_fall = Signal(8)
        self.comb += self.di.eq(Cat(di_fall, di_rise))

        # OPI DDR registers
        dq_delayed = Signal(8)
        self.specials += dq.get_tristate(pads.dq[1:])
        for i in range(1, 8):
            self.specials += Instance("ODDR",
                p_DDR_CLK_EDGE = "SAME_EDGE",
                i_C  = ClockSignal(),
                i_R  = ResetSignal(),
                i_S  = 0,
                i_CE = 1,
                i_D1 = do_rise[i],
                i_D2 = do_fall[i],
                o_Q  = dq.o[i-1],
            )
            if i == 1: # Only wire up o_CNTVALUEOUT for one instance
                self.specials += Instance(idelay_name,
                    p_DELAY_SRC             = "IDATAIN",
                    p_SIGNAL_PATTERN        = "DATA",
                    p_CINVCTRL_SEL          = "FALSE",
                    p_HIGH_PERFORMANCE_MODE = "FALSE",
                    p_REFCLK_FREQUENCY      = 200.0,
                    p_PIPE_SEL              = "FALSE",
                    p_IDELAY_VALUE          = dq_delay_taps,
                    p_IDELAY_TYPE           = delay_type,

                    i_C           = ClockSignal(),
                    i_CINVCTRL    = 0,
                    i_REGRST      = 0,
                    i_LDPIPEEN    = 0,
                    i_INC         = 0,
                    i_CE          = 0,
                    i_LD          = self.delay_update,
                    i_CNTVALUEIN  = self.delay_config.fields.d,
                    o_CNTVALUEOUT = self.delay_status.fields.q,
                    i_IDATAIN     = dq.i[i-1],
                    o_DATAOUT     = dq_delayed[i],
                    i_DATAIN=0,
                ),
            else: # Don't wire up o_CNTVALUEOUT for others
                self.specials += Instance(idelay_name,
                    p_DELAY_SRC             = "IDATAIN",
                    p_SIGNAL_PATTERN        = "DATA",
                    p_CINVCTRL_SEL          = "FALSE",
                    p_HIGH_PERFORMANCE_MODE = "FALSE",
                    p_REFCLK_FREQUENCY      = 200.0,
                    p_PIPE_SEL              = "FALSE",
                    p_IDELAY_VALUE          = dq_delay_taps,
                    p_IDELAY_TYPE           = delay_type,
                    i_C          = ClockSignal(),
                    i_CINVCTRL   = 0,
                    i_REGRST     = 0,
                    i_LDPIPEEN   = 0 ,
                    i_INC        = 0,
                    i_CE         = 0,
                    i_LD         = self.delay_update,
                    i_CNTVALUEIN = self.delay_config.fields.d,
                    i_IDATAIN    = dq.i[i-1],
                    o_DATAOUT    = dq_delayed[i],
                    i_DATAIN=0,
                ),

            self.specials += Instance("IDDR", name="{}{}".format(iddr_name, str(i)),
                p_DDR_CLK_EDGE = "SAME_EDGE_PIPELINED",
                i_C  = dqs_iobuf,
                i_R  = ResetSignal(),
                i_S  = 0,
                i_CE = 1,
                i_D  = dq_delayed[i],
                o_Q1 = di_rise[i],
                o_Q2 = di_fall[i],
            )
        # SPI SDR register
        self.specials += [
            Instance("FDRE", name="{}".format(cipo_name),
                i_C  = ~ClockSignal("spinor"),
                i_CE = 1,
                i_R  = 0,
                o_Q  = self.cipo,
                i_D  = dq_delayed[1],
            )
        ]

        # bit 0 (copi) is special-cased to handle SPI mode
        self.specials += dq_copi.get_tristate(pads.dq[0])
        do_mux_rise = Signal() # mux signal for copi/dq select of bit 0
        do_mux_fall = Signal()
        self.specials += [
            Instance("ODDR",
                p_DDR_CLK_EDGE = "SAME_EDGE",
                i_C  = ClockSignal(),
                i_R  = ResetSignal(),
                i_S  = 0,
                i_CE = 1,
                i_D1 = do_mux_rise,
                i_D2 = do_mux_fall,
                o_Q  = dq_copi.o,
            ),
            Instance("IDDR",
                p_DDR_CLK_EDGE="SAME_EDGE_PIPELINED",
                i_C  = dqs_iobuf,
                i_R  = ResetSignal(),
                i_S  = 0,
                i_CE = 1,
                o_Q1 = di_rise[0],
                o_Q2 = di_fall[0],
                i_D  = dq_delayed[0],
            ),
        ]
        self.specials += Instance(idelay_name,
            p_DELAY_SRC             = "IDATAIN",
            p_SIGNAL_PATTERN        = "DATA",
            p_CINVCTRL_SEL          = "FALSE",
            p_HIGH_PERFORMANCE_MODE = "FALSE",
            p_REFCLK_FREQUENCY      = 200.0,
            p_PIPE_SEL              = "FALSE",
            p_IDELAY_VALUE          = dq_delay_taps,
            p_IDELAY_TYPE           = delay_type,

            i_C          = ClockSignal(),
            i_CINVCTRL   = 0,
            i_REGRST     = 0,
            i_LDPIPEEN   = 0,
            i_INC        = 0,
            i_CE         = 0,
            i_LD         = self.delay_update,
            i_CNTVALUEIN = self.delay_config.fields.d,
            i_IDATAIN    = dq_copi.i,
            o_DATAOUT    = dq_delayed[0],
            i_DATAIN=0,
        ),

        # Wire up SCLK interface
        clk_en = Signal()
        self.specials += [
            # De-activate the CCLK interface, parallel it with a GPIO
            Instance("STARTUPE2",
                i_CLK       = 0,
                i_GSR       = self.gsr,
                i_GTS       = 0,
                i_KEYCLEARB = self.keyclearb,
                i_PACK      = 0,
                i_USRDONEO  = 1,
                i_USRDONETS = 1,
                i_USRCCLKO  = 0,
                i_USRCCLKTS = 1,  # Force to tristate
                o_CFGMCLK   = self.cfgmclk,
            ),
            Instance("ODDR", name=sclk_name, # Need to name this so we can constrain it properly
                p_DDR_CLK_EDGE = "SAME_EDGE",
                i_C  = ClockSignal("spinor"),
                i_R  = ResetSignal("spinor"),
                i_S  = 0,
                i_CE = 1,
                i_D1 = clk_en,
                i_D2 = 0,
                o_Q  = pads.sclk,
            )
        ]

        # wire up CS_N
        spi_cs_n = Signal()
        opi_cs_n = Signal()
        self.comb += cs_n.eq( (spi_mode & spi_cs_n) | (~spi_mode & opi_cs_n) )
        self.specials += [
            Instance("ODDR",
              p_DDR_CLK_EDGE="SAME_EDGE",
              i_C=ClockSignal(), i_R=0, i_S=ResetSignal(), i_CE=1,
              i_D1=cs_n, i_D2=cs_n, o_Q=pads.cs_n,
            ),
        ]

        self.architecture = ModuleDoc("""Architecture

        The machine is split into two separate pieces, one to handle SPI, and one to handle OPI.

        SPI
        -----

        The SPI machine architecture is split into two levels: MAC and PHY.

        The MAC layer is responsible for:
        - receiving requests via CSR register to perform config/status/special command sequences,
        and dispatching these to the SPI PHY
        - translating wishbone bus requests into command sequences, and routing them to either OPI
        or SPI PHY.
        - managing the chip select to the chip, and ensuring that one dummy cycle is inserted after
        chip select is asserted, or before it is de-asserted; and that the chip select "high" times
        are adequate (1 cycle between reads, 4 cycles for all other operations)

        On boot, the interface runs in SPI; once the wakeup sequence is executed, the chip permanently
        switches to OPI mode unless the CR2 registers are written to fall back, or the
        reset to the chip is asserted.

        The PHY layers are responsible for the following tasks:
        - Serializing and deserializing data, standardized on 8 bits for SPI and 16 bits for OPI
        - counting dummy cycles
        - managing the clock enable

        PHY cycles are initiated with a "req" signal, which is only sampled for
        one cycle and then ignored until the PHY issues an "ack" that the current cycle is complete.
        Thus holding "req" high can allow the PHY to back-to-back issue cycles without pause.

        OPI
        -----

        The OPI machine is split into three parts: a command controller, a Tx PHY, and an Rx PHY.

        The Tx PHY is configured with a "dummy cycle" count register, as there is a variable length
        delay for dummy cycles in OPI.

        In OPI mode, read data is `mesochronous`, that is, they return at precisely the same frequency
        as SCLK, but with an unknown phase relationship. The DQS strobe is provided as a "hint" to
        the receiving side to help retime the data. The mesochronous nature of the read data is why
        the Tx and Rx PHY must be split into two separate machines, as they are operating in
        different clock domains.

        DQS is implemented on the ROM as an extra data output that is guaranteed to change polarity
        with each data byte; the skew mismatch of DQS to data is within +/-0.6ns or so. It turns out
        the mere act of routing the DQS into a BUFR buffer before clocking the data into an IDDR
        primitive is sufficient to delay the DQS signal and meet setup and hold time on the IDDR.

        Once captured by the IDDR, the data is fed into a dual-clock FIFO to make the transition
        from the DQS to sysclk domains cleanly.

        Because of the latency involved in going from pin->IDDR->FIFO, excess read cycles are
        required beyond the end of the requested cache line. However, there is virtually no penalty
        in pre-filling the FIFO with data; if a new cache line has to be fetched, the FIFO can simply
        be reset and all pointers zeroed. In fact, pre-filling the FIFO can lead to great performance
        benefits if sequential cache lines are requested. In simulation, a cache line can be filled
        in 10 bus cycles if it happens to be prefetched (as opposed to 49 bus cycles for random reads).
        Either way, this compares favorably to 288 cycles for random reads in 100MHz SPI mode (or 576
        for the spimemio.v, which runs at 50MHz).

        The command controller is repsonsible for sequencing all commands other than fast reads. Most
        commands have some special-case structure to them, and as more commands are implemented, the
        state machine is expected to grow fairly large. Fast reads are directly handled in "tx_run"
        mode, where the TxPhy and RxPhy run a tight loop to watch incoming read bus cycles, check
        the current address, fill the prefetch fifo, and respond to bus cycles.

        Writes to ROM might lock up the machine; a TODO is to test this and do something more sane,
        like ignore writes by sending an ACK immediately while discarding the data.

        Thus, an OPI read proceeds as follows:

        - When BUS/STB are asserted:
           TxPhy:

           - capture bus_adr, and compare against the *next read* address pointer
              - if they match, allow the PHYs to do the work

           - if bus_adr and next read address don't match, save to next read address pointer, and
             cycle wr/rd clk for 5 cycle while asserting reset to reset the FIFO
           - initiate an 8DTRD with the read address pointer
           - wait the specified dummy cycles

           - greedily pre-fill the FIFO by continuing to clock DQS until either:
             - the FIFO is full
             - pre-fetch is aborted because bus_adr and next read address don't match and FIFO is reset

           RxPHY:

           - while CTI==2, assemble data into 32-bit words as soon as EMPTY is deasserted,
             present a bus_ack, and increment the next read address pointer
           - when CTI==7, ack the data, and wait until the next bus cycle with CTI==2 to resume
             reading

        - A FIFO_SYNC_MACRO is used to instantiate the FIFO. This is chosen because:
           - we can specify RAMB18's, which seem to be under-utilized by the auto-inferred memories by migen
           - the XPM_FIFO_ASYNC macro claims no instantiation support, and also looks like it has weird
             requirements for resetting the pointers: you must check the reset outputs, and the time to
             reset is reported to be as high as around 200ns (anecdotally -- could be just that the sim I
             read on the web is using a really slow clock, but I'm guessing it's around 10 cycles).
           - the FIFO_SYNC_MACRO has a well-specified fixed reset latency of 5 cycles.
           - The main downside of FIFO_SYNC_MACRO over XPM_FIFO_ASYNC is that XPM_FIFO_ASYNC can automatically
             allow for output data to be read at 32-bit widths, with writes at 16-bit widths. However, with a
             bit of additional logic and pipelining, we can aggregate data into 32-bit words going into a
             32-bit FIFO_SYNC_MACRO, which is what we do in this implementation.
        """)
        self.bus = bus = wishbone.Interface()

        self.command = CSRStorage(description="Write individual bits to issue special commands to SPI; setting multiple bits at once leads to undefined behavior.",
            fields=[
                CSRField("wakeup",       size=1, description="Sequence through init & wakeup routine"),
                CSRField("exec_cmd",     size=1, description="Writing a `1` executes a manual command", pulse=True),
                CSRField("cmd_code",     size=8, description="Manual command code (first 8 bits, e.g. PP4B is 0x12)"),
                CSRField("has_arg",      size=1, description="When set, transmits the value of `cmd_arg` as the argument to the command"),
                # CSRField("write_cmd", size=1, description="When `1`, `data_bytes` are written from page FIFO; when `0`, up to 4 STR `data_bytes` are read into readback CSR"),
                CSRField("dummy_cycles", size=5, description="Number of dummy cycles for manual command; 0 implies a write, >0 implies read"),
                CSRField("data_words",   size=8, description="Number of data words (2x bytes)"),
                CSRField("lock_reads",   size=1, description="When set, locks out read operations (recommended when doing programming)"),
            ])
        self.cmd_arg = CSRStorage(description="Command argument",
            fields=[
                CSRField("cmd_arg", size=32, description="Argument to manual command")
            ])
        self.cmd_rbk_data = CSRStatus(description = "Readback data from commands",
            fields=[
                CSRField("cmd_rbk_data", size=32, description="Data read back from a cmd_code that has `write_code` set to 0"),
            ]
        )
        self.status = CSRStatus(description="Interface status",
            fields=[
                CSRField("wip", size=1, description="Operation in progress (write or erease)")
            ])
        self.wdata = CSRStorage(description="Page data to write to FLASH",
            fields = [
                CSRField("wdata", size=16, description="""16-bit wide write data presented to FLASH, committed to a 128-entry deep FIFO.
                Writes to this register are not cached; note that writes to the SPINOR address space are also committed
                to the FIFO, but this space is cached by the CPU, and therefore not guaranteed to be coherent or in order.
                The direct wishbone-write address space is provisioned for e.g. USB bus masters that don't have caching.""")
            ]
        )
        # TODO: implement ECC detailed register readback, CRC checking

        # PHY machine mux --------------------------------------------------------------------------
        # clk_en mux
        spi_clk_en = Signal()
        opi_clk_en = Signal()
        self.sync += clk_en.eq(~spi_mode & opi_clk_en | spi_mode & spi_clk_en)
        # Tristate mux
        self.sync += [
            dq.oe.eq(~spi_mode & self.tx),
            dq_copi.oe.eq(spi_mode | self.tx),
        ]
        # Data out mux (no data in mux, as we can just sample data in all the time without harm)
        self.comb += do_mux_rise.eq(~spi_mode & do_rise[0] | spi_mode & self.copi)
        self.comb += do_mux_fall.eq(~spi_mode & do_fall[0] | spi_mode & self.copi)

        # Indicates if the current "req" requires dummy cycles to be appended (used for both OPI/SPI)
        has_dummy = Signal()
        # Location of the internal ROM address pointer; reset to invalid address to force an address
        # request on first read
        rom_addr  = Signal(32, reset=0xFFFFFFFC)

        # MAC/PHY abstraction for OPI
        txphy_do = Signal(16) # Two sources of data out for OPI, one from the PHY, one from MAC
        txcmd_do = Signal(16)
        opi_di   = Signal(16)

        # Internal machines
        opi_addr         = Signal(32)
        opi_fifo_rd      = Signal(32)
        opi_fifo_wd      = Signal(32)
        opi_reset_rx_req = Signal()
        opi_reset_rx_ack = Signal()
        opi_rx_run       = Signal()

        rx_almostempty = Signal()
        rx_almostfull  = Signal()
        rx_empty       = Signal()
        rx_full        = Signal()
        rx_rdcount     = Signal(9)
        rx_rderr       = Signal()
        rx_wrcount     = Signal(9)
        rx_wrerr       = Signal()
        rx_rden        = Signal()
        rx_wren        = Signal(reset=1)
        rx_fifo_rst    = Signal()

        wrendiv  = Signal()
        wrendiv2 = Signal()
        rx_fifo_rst_pipe = Signal()
        cmd_run = Signal()
        self.specials += [
            # This next pair of async-clear flip flops creates a write-enable gate that (a) ignores
            # the first two DQS strobes (as they are pipe-filling) and (b) alternates with the correct
            # phase so we are sampling 32-bit data into the FIFO.
            Instance("FDCE", name="FDCE_WREN",
                i_C   = dqs_iobuf,
                i_D   = ~wrendiv,
                o_Q   = wrendiv,
                i_CE  = 1,
                i_CLR = ~(rx_wren & ~cmd_run),
            ),
            Instance("FDCE", name="FDCE_WREN",
                i_C   = dqs_iobuf,
                i_D   = ~wrendiv2,
                o_Q   = wrendiv2,
                i_CE  = wrendiv & ~wrendiv2,
                i_CLR = ~(rx_wren & ~cmd_run),
            ),
            # Direct FIFO primitive is more resource-efficient and faster than migen primitive.
            Instance("FIFO_DUALCLOCK_MACRO",
                p_DEVICE                  = "7SERIES",
                p_FIFO_SIZE               = "18Kb",
                p_DATA_WIDTH              = 32,
                p_FIRST_WORD_FALL_THROUGH = "TRUE",
                p_ALMOST_EMPTY_OFFSET     = 6,
                p_ALMOST_FULL_OFFSET      = (511 - (8*prefetch_lines + 8)),  # a few extra entries needed to meet DRC...

                o_ALMOSTEMPTY = rx_almostempty,
                o_ALMOSTFULL  = rx_almostfull,
                o_DO          = opi_fifo_rd,
                o_EMPTY       = rx_empty,
                o_FULL        = rx_full,
                o_RDCOUNT     = rx_rdcount,
                o_RDERR       = rx_rderr,
                o_WRCOUNT     = rx_wrcount,
                o_WRERR       = rx_wrerr,
                i_DI          = opi_fifo_wd,
                i_RDCLK       = ClockSignal(),
                i_RDEN        = rx_rden,
                i_WRCLK       = dqs_iobuf,
                i_WREN        = wrendiv & wrendiv2,
                i_RST         = rx_fifo_rst_pipe, #rx_fifo_rst,
            )
        ]
        self.sync.dqs += opi_di.eq(self.di)
        self.comb += opi_fifo_wd.eq(Cat(opi_di, self.di))
        self.sync += rx_fifo_rst_pipe.eq(rx_fifo_rst) # add one pipe register to help relax this timing path. It is critical so it must be timed, but one extra cycle is OK.
        rbk_data = Signal(32)
        self.sync += rbk_data.eq(opi_fifo_wd) # buffer for capture to CSR on command cycles
        bus_ack_r = Signal()
        bus_ack_w = Signal()

        #---------  Page write data responder -----------------------
        self.submodules.txwr_fifo = SyncFIFOBuffered(width=16, depth=128)
        self.submodules.pgwr = pgwr = FSM(reset_state="IDLE")
        pgwr.act("IDLE",
            If(self.wdata.re,
                self.txwr_fifo.we.eq(1),
                self.txwr_fifo.din.eq(self.wdata.fields.wdata)
            ).Elif(bus.cyc & bus.stb & bus.we,
                self.txwr_fifo.din.eq(bus.dat_w[:16]),  # lower 16 bits first
                self.txwr_fifo.we.eq(1),
                NextState("HIWORD")
            ).Else(
                self.txwr_fifo.we.eq(0),
                bus_ack_w.eq(0)
            )
        )
        pgwr.act("HIWORD",
            self.txwr_fifo.din.eq(bus.dat_w[16:]),  # top 16 next
            self.txwr_fifo.we.eq(1),
            bus_ack_w.eq(1),
            NextState("WAIT_DONE")
        )
        pgwr.act("WAIT_DONE",
            If( ~(bus.cyc & bus.stb & bus.we),
                NextState("IDLE"),
                bus_ack_w.eq(0),
            ).Else(
                bus_ack_w.eq(1),
            )
        )
        self.comb += [
            bus.ack.eq(bus_ack_r | bus_ack_w),
        ]

        #---------  OPI Rx Phy machine ------------------------------
        self.submodules.rxphy = rxphy = FSM(reset_state="IDLE")
        rxphy_cnt = Signal(3)
        rxphy.act("IDLE",
            If(spi_mode,
                NextState("IDLE"),
            ).Else(
                NextValue(bus_ack_r, 0),
                If(opi_reset_rx_req,
                    NextState("WAIT_RESET"),
                    NextValue(rxphy_cnt, 6),
                    NextValue(rx_wren, 0),
                    NextValue(rx_fifo_rst, 1)
                ).Elif(opi_rx_run,
                    NextValue(rx_wren, 1),
                    If((bus.cyc & bus.stb & ~bus.we) & ((bus.cti == 2) | (bus.cti == 0) |
                       ((bus.cti == 7) & ~bus.ack) ), # handle case of non-pipelined read, ack is late
                        If(~rx_empty,
                            NextValue(bus.dat_r, opi_fifo_rd),
                            rx_rden.eq(1),
                            NextValue(opi_addr, opi_addr + 4),
                            NextValue(bus_ack_r, 1)
                        )
                    )
                )
            )
        )
        rxphy.act("WAIT_RESET",
            NextValue(opi_addr, Cat(Signal(2), bus.adr)),
            NextValue(rxphy_cnt, rxphy_cnt - 1),
            If(rxphy_cnt == 0,
                NextValue(rx_fifo_rst, 0),
                opi_reset_rx_ack.eq(1),
                NextState("IDLE")
            )
        )


        # TxPHY machine: OPI -------------------------------------------------------------------------
        run_is_hot = Signal() # indicates that the receive FIFO is hot and needs a reset before going into a cmd
        cmd_req = Signal()
        cmd_ack = Signal()
        self.sync += [
            If(self.command.fields.exec_cmd,
                cmd_req.eq(1),
            ).Elif(cmd_ack,
                cmd_req.eq(0),
            ).Else(
                cmd_req.eq(cmd_req)
            )
        ]
        cmd_done = Signal()
        wip_state = Signal()
        self.comb += self.status.fields.wip.eq(wip_state | cmd_req) # need combinational loop-back to repsond to fast WIP inquiries
        self.sync += [
            If(cmd_done,
                wip_state.eq(0),
            ).Elif(cmd_run | cmd_req | cmd_done, # lock out writing through the entire life cycle
                wip_state.eq(1)
            ).Else(
                wip_state.eq(wip_state)
            )
        ]

        txphy_cnt   = Signal(4)
        tx_run      = Signal()
        txphy_cs_n  = Signal(reset=1)
        txcmd_cs_n  = Signal(reset=1)
        txphy_clken = Signal()
        txcmd_clken = Signal()
        txphy_oe    = Signal()
        txcmd_oe    = Signal()
        txwr_cnt = Signal(8)
        tx_run_d = Signal()
        self.sync += tx_run_d.eq(tx_run)
        self.sync += opi_cs_n.eq( (tx_run_d & txphy_cs_n) | (~tx_run_d & ~cmd_run & txcmd_cs_n) | (cmd_run & txphy_cs_n) )
        self.comb += If( tx_run | cmd_run, self.do.eq(txphy_do) ).Else( self.do.eq(txcmd_do) )
        self.comb += opi_clk_en.eq( (tx_run & txphy_clken) | (~tx_run & txcmd_clken) | (cmd_run & txphy_clken) )
        self.comb += self.tx.eq( (tx_run & txphy_oe) | (~tx_run & txcmd_oe) | (cmd_run & txphy_oe) )
        tx_almostfull = Signal()
        self.sync += tx_almostfull.eq(rx_almostfull) # sync the rx_almostfull signal into the local clock domain
        txphy_bus = Signal()
        self.sync += txphy_bus.eq(bus.cyc & bus.stb & ~bus.we & ((bus.cti == 2) | (bus.cti == 0)))
        tx_resetcycle = Signal()

        self.submodules.txphy = txphy = FSM(reset_state="RESET")
        txphy.act("RESET",
            NextValue(opi_rx_run, 0),
            NextValue(txphy_oe, 0),
            NextValue(txphy_cs_n, 1),
            NextValue(txphy_clken, 0),
            NextValue(cmd_done, 0),
            # guarantee that the first state we go to out of reset is a four-cycle burst
            NextValue(txphy_cnt, 4),
            If( tx_run & ~spi_mode,
                NextState("TX_SETUP")
            ).Elif( cmd_run & ~spi_mode & ~cmd_done,  # have to look at cmd_done because of delay from done-to-clear of run
                If(run_is_hot,
                    NextValue(txphy_clken, 1),
                    NextValue(opi_reset_rx_req, 1),
                    NextValue(txphy_cs_n, 0),
                    NextState("TX_RESET_BEFORE_CMD"),
                ).Else(
                    NextState("TX_SETUP_CMD")
                )
            )
        )
        txphy.act("TX_SETUP",
            NextValue(opi_rx_run, 0),
            NextValue(txphy_cnt, txphy_cnt - 1),
            If( txphy_cnt > 0,
                NextValue(txphy_cs_n, 1)
            ).Else(
                NextValue(txphy_cs_n, 0),
                NextValue(txphy_oe, 1),
                NextState("TX_CMD_CS_DELAY")
            )
        )
        txphy.act("TX_CMD_CS_DELAY",  # meet setup timing for CS-to-clock
            If( tx_run,
               NextState("TX_CMD")
            )
        )
        txphy.act("TX_CMD",
            NextValue(txphy_do, 0xEE11),
            NextValue(txphy_clken, 1),
            NextState("TX_ADRHI")
        )
        txphy.act("TX_ADRHI",
            NextValue(txphy_do, opi_addr[16:] & 0x07FF), # mask off unused bits
            NextState("TX_ADRLO")
        )
        txphy.act("TX_ADRLO",
                  NextValue(txphy_do, opi_addr[:16]),
                  NextValue(txphy_cnt, self.config.fields.dummy - 1),
                  NextState("TX_DUMMY")
        )
        txphy.act("TX_DUMMY",
            NextValue(txphy_oe, 0),
            NextValue(txphy_do, 0),
            NextValue(txphy_cnt, txphy_cnt - 1),
            If(txphy_cnt == 0,
                NextValue(opi_rx_run, 1),
                If(tx_resetcycle,
                    NextValue(txphy_clken, 1),
                    NextValue(opi_reset_rx_req, 1),
                    NextState("TX_RESET_RX"),
                ).Else(
                    NextState("TX_FILL"),
               )
            )
        )
        txphy.act("TX_FILL",
            If(tx_run,
                If(((~txphy_bus & (bus.cyc & bus.stb & ~bus.we & ((bus.cti == 2) | (bus.cti == 0)) )) &
                   (opi_addr[2:] != bus.adr)) | tx_resetcycle,
                    # Tt's a new bus cycle, and the requested address is not equal to the current
                    # read buffer address
                    NextValue(txphy_clken, 1),
                    NextValue(opi_reset_rx_req, 1),
                    NextState("TX_RESET_RX"),
               ).Else(
                    If(tx_almostfull & ~bus.ack,
                        NextValue(txphy_clken, 0)
                    ).Else(
                        NextValue(txphy_clken, 1)
                    )
               ),
               If(~(bus.cyc & bus.stb),
                    NextValue(opi_rx_run, 0),
               ).Else(
                    NextValue(opi_rx_run, 1),
               )
            ).Else(
                NextValue(txphy_clken, 0),
                NextState("RESET")
            )
        )
        txphy.act("TX_RESET_RX", # Keep clocking the RX until it acknowledges a reset
            NextValue(opi_rx_run, 0),
            NextValue(opi_reset_rx_req, 0),
            If(opi_reset_rx_ack,
                NextValue(txphy_clken, 0),
                NextValue(txphy_cnt, 0), # 1 cycle CS on back-to-back reads
                NextValue(txphy_cs_n, 1),
                NextState("TX_SETUP"),
            ).Else(
                NextValue(txphy_clken, 1),
            )
        )
        # issue a Rx FIFO reset before going into command mode
        txphy.act("TX_RESET_BEFORE_CMD",
            NextValue(txphy_clken, 1),
            NextValue(opi_reset_rx_req, 0),
            If(opi_reset_rx_ack,
                NextValue(txphy_clken, 0),
                NextState("TX_SETUP_CMD")
            )
        )
        # mirror setup here because once we count down the delay, it must be atomic to this FSM path
        # and we need the full 40ns of CS delay every time we go down this path!
        txphy.act("TX_SETUP_CMD",
            NextValue(opi_rx_run, 0),
            NextValue(txphy_cnt, txphy_cnt - 1),
            If( txphy_cnt > 0,
                NextValue(txphy_cs_n, 1)
            ).Else(
                NextValue(txphy_cs_n, 0),
                NextValue(txphy_oe, 1),
                NextState("TX_CMD_MAN_CS_DELAY")
            )
        )
        txphy.act("TX_CMD_MAN_CS_DELAY",
            NextState("TX_MAN_CMD")
        ),
        txphy.act("TX_MAN_CMD",
            NextValue(txphy_do, Cat(~self.command.fields.cmd_code, self.command.fields.cmd_code)),
            NextValue(txphy_clken, 1),
            If(self.command.fields.has_arg,
                NextState("TX_ARGHI")
            ).Elif(self.command.fields.dummy_cycles > 0, # implies a read
                NextValue(txphy_cnt, self.command.fields.dummy_cycles - 1),
                NextState("TX_MAN_DUMMY")
            ).Elif(self.command.fields.data_words > 0,  # write is implied if dummy cycles is 0
                NextValue(txwr_cnt, self.command.fields.data_words),
                NextState("TX_WRDATA")
            ).Else( # simple command with no data or readback
                NextState("TX_WR_RESET"),
            )
        )
        txphy.act("TX_ARGHI",
            NextValue(txphy_do, self.cmd_arg.fields.cmd_arg[16:]),
            NextState("TX_ARGLO")
        )
        txphy.act("TX_ARGLO",
            NextValue(txphy_do, self.cmd_arg.fields.cmd_arg[:16]),
            If(self.command.fields.dummy_cycles > 0,
               NextValue(txphy_cnt, self.command.fields.dummy_cycles - 1),
               NextState("TX_MAN_DUMMY")
            ).Elif(self.command.fields.data_words > 0, # self.command.fields.write_cmd,  # write is implied if dummy cycles is 0 and data exists
                NextValue(txwr_cnt, self.command.fields.data_words - 1),
                NextState("TX_WRDATA")
            ).Else(
                NextState("TX_WR_RESET")
            )
        )
        txphy.act("TX_MAN_DUMMY",
            NextValue(txphy_oe, 0),
            NextValue(txphy_do, 0),
            NextValue(txphy_cnt, txphy_cnt - 1),
            If(txphy_cnt == 0,
                # always a readback after a dummy cycle
                # ignore upper bits, and note that +1 cycle is added because we have to pump DQS a dummy cycle to push data through the rbk pipe
                # the SEEPROM mostly handles the extra pump OK.
                NextValue(txphy_cnt, self.command.fields.data_words[:4] - 1 + 1),
                NextState("TX_MAN_RBK"),
            )
        )
        txphy.act("TX_MAN_RBK",
            If(txphy_cnt == 0,
                NextState("TX_MAN_RBK_WAIT"),
                NextValue(txphy_cnt, 4),
            ).Else(
                NextValue(txphy_cnt, txphy_cnt - 1),
            )
        )
        txphy.act("TX_MAN_RBK_WAIT", # need to wait some cycles for the readback data to return from the device before latching it
            If(txphy_cnt == 0,
                NextValue(self.cmd_rbk_data.fields.cmd_rbk_data, rbk_data),
                NextState("TX_WR_RESET"),
            ).Else(
                NextValue(txphy_cnt, txphy_cnt - 1),
            )
        )
        txphy.act("TX_WRDATA",
            If(txwr_cnt == 0,
                NextValue(txphy_do, self.txwr_fifo.dout),
                NextState("TX_WR_RESET"),
            ).Else(
                NextValue(txwr_cnt, txwr_cnt - 1),
                NextValue(txphy_do, self.txwr_fifo.dout),
                self.txwr_fifo.re.eq(1),
            )
        )
        txphy.act("TX_WR_RESET",
            NextValue(txphy_oe, 0),
            NextValue(txphy_clken, 0),
            # drain any excess values in the page FIFO
            If(self.txwr_fifo.readable,
                self.txwr_fifo.re.eq(1),
            ).Else(
                NextState("RESET"),
                NextValue(cmd_done, 1),
            )
        )


        #---------  OPI CMD machine ------------------------------
        self.submodules.opicmd = opicmd = FSM(reset_state="RESET")
        opicmd.act("RESET",
            NextValue(txcmd_do, 0),
            NextValue(txcmd_oe, 0),
            NextValue(tx_run, 0),
            NextValue(cmd_run, 0),
            NextValue(txcmd_cs_n, 1),
            NextValue(run_is_hot, 0),
            If(~spi_mode,
                NextState("IDLE")
            ).Else(
                NextState("RESET_CYCLE")
            ),
        )
        opicmd.act("RESET_CYCLE",
            NextValue(txcmd_cs_n, 0),
            If(opi_reset_rx_ack,
                NextValue(tx_run, 1),
                NextState("IDLE"),
            ).Else(
                NextValue(tx_run, 1),
                tx_resetcycle.eq(1)
            )
        )
        opicmd.act("IDLE",
            NextValue(txcmd_cs_n, 1),
            If(~spi_mode,  # This machine stays in idle once spi_mode is dropped
               ## The full form of this machine is as follows:
               # - First check if there is a CSR special command pending
               #   - if so, wait until the current bus cycle is done, then de-assert tx_run
               #   - then run the command
               # - Else wait until a bus cycle, and once it happens, put the system into run mode
               If(bus.cyc & bus.stb & ~self.command.fields.lock_reads,
                    If(~bus.we & ((bus.cti == 2) | (bus.cti == 0)),
                        If(~run_is_hot,
                            NextValue(opi_addr, Cat(Signal(2), bus.adr)),
                        ),
                        NextValue(run_is_hot, 1),
                        NextState("TX_RUN")
                    ).Else(
                        # Handle other cases here, e.g. what do we do if we get a write? probably
                        # should just ACK it without doing anything so the CPU doesn't freeze...
                    )
               ).Elif(cmd_req,
                    NextState("DISPATCH_CMD"),
               )
            )
        )
        opicmd.act("TX_RUN",
            NextValue(tx_run, 1),
            If(cmd_req | self.command.fields.lock_reads, # Respond to commands
                NextState("WAIT_DISPATCH")
            )
        )
        # Wait until the current cycle is done, then stop TX and dispatch command
        opicmd.act("WAIT_DISPATCH",
            If( ~(bus.cyc & bus.stb),
                NextValue(tx_run, 0),
                NextValue(cmd_run, 1),
                NextState("DISPATCH_CMD")
            )
        )
        opicmd.act("DISPATCH_CMD",
            cmd_ack.eq(1), # clear the command dispatch pulse cache
            If(cmd_done,
                NextValue(run_is_hot, 0),
                NextValue(cmd_run, 0),
                NextValue(tx_run, 0),
                NextState("IDLE"),
            ).Else(
                NextValue(cmd_run, 1),
            )
        )

        ############################################################################################
        ############################################################################################
        ############################################################################################
        ############################################################################################
        # MAC/PHY abstraction for the SPI machine
        spi_req = Signal()
        spi_ack = Signal()
        spi_do  = Signal(8) # this is the API to the machine
        spi_di  = Signal(8)

        # PHY machine: SPI -------------------------------------------------------------------------

        # internal signals are:
        # selection - spi_mode
        # OPI - self.do(16), self.di(16), self.tx
        # SPI - self.copi, self.cipo
        # cs_n - both
        # ecs_n - OPI
        # clk_en - both

        spicount     = Signal(5)
        spi_so       = Signal(8) # this internal to the machine
        spi_si       = Signal(8)
        spi_dummy    = Signal()
        spi_di_load  = Signal() # spi_do load is pipelined back one cycle using this mechanism
        spi_di_load2 = Signal()
        spi_ack_pipe = Signal()
        # Pipelining is required the cipo path is very slow (IOB->fabric FD), and a falling-edge
        # retiming reg is used to meet timing
        self.sync += [
            spi_di_load2.eq(spi_di_load),
            If(spi_di_load2, spi_di.eq(Cat(self.cipo, spi_si[:-1]))).Else(spi_di.eq(spi_di)),
            spi_ack.eq(spi_ack_pipe),
        ]
        self.comb += self.copi.eq(spi_so[7])
        self.sync += spi_si.eq(Cat(self.cipo, spi_si[:-1]))
        self.submodules.spiphy = spiphy = FSM(reset_state="RESET")
        spiphy.act("RESET",
            If(spi_req,
                NextState("REQ"),
                NextValue(spicount, 7),
                NextValue(spi_clk_en, 1),
                NextValue(spi_so, spi_do),
                NextValue(spi_dummy, has_dummy),
            ).Else(
                NextValue(spi_clk_en, 0),
                NextValue(spi_ack_pipe, 0),
                NextValue(spicount, 0),
                NextValue(spi_dummy, 0),
            )
        )
        spiphy.act("REQ",
            If(spicount > 0,
                NextValue(spicount, spicount-1),
                NextValue(spi_clk_en, 1),
                NextValue(spi_so, Cat(0, spi_so[:-1])),
                NextValue(spi_ack_pipe, 0)
            ).Elif( (spicount == 0) & spi_req & ~spi_dummy, # Back-to-back transaction
                NextValue(spi_clk_en, 1),
                NextValue(spicount, 7),
                NextValue(spi_clk_en, 1),
                NextValue(spi_so, spi_do), # Reload the so register
                spi_di_load.eq(1), # "naked" .eq() create single-cycle pulses that default back to 0
                NextValue(spi_ack_pipe, 1),
                NextValue(spi_dummy, has_dummy)
            ).Elif( (spicount == 0) & ~spi_req & ~spi_dummy, # Go back to idle
                spi_di_load.eq(1),
                NextValue(spi_ack_pipe, 1),
                NextValue(spi_clk_en, 0),
                NextState("RESET")
            ).Elif( (spicount == 0) & spi_dummy,
                spi_di_load.eq(1),
                NextValue(spicount, self.config.fields.dummy),
                NextValue(spi_clk_en, 1),
                NextValue(spi_ack_pipe, 0),
                NextValue(spi_so, 0),  # Do a dummy with '0' as the output
                NextState("DUMMY")
            ) # This actually should be a fully defined situation, no "Else" applicable
        )
        spiphy.act("DUMMY",
            If(spicount > 1, # Instead of doing dummy-1, we stop at count == 1
                NextValue(spicount, spicount - 1),
                NextValue(spi_clk_en, 1)
            ).Elif(spicount <= 1 & spi_req,
                NextValue(spi_clk_en, 1),
                NextValue(spicount, 7),
                NextValue(spi_so, spi_do),  # Reload the so register
                NextValue(spi_ack_pipe, 1), # Finally ack the cycle
                NextValue(spi_dummy, has_dummy)
            ).Else(
                NextValue(spi_clk_en, 0),
                NextValue(spi_ack_pipe, 1),  # Finally ack the cycle
                NextState("RESET")
            )
        )

        # SPI MAC machine --------------------------------------------------------------------------
        # default active on boot
        addr_updated = Signal()
        d_to_wb      = Signal(32) # data going back to wishbone
        mac_count    = Signal(5)
        new_cycle    = Signal(1)
        self.submodules.mac = mac = FSM(reset_state="RESET")
        mac.act("RESET",
                NextValue(spi_mode, 1),
                NextValue(addr_updated, 0),
                NextValue(d_to_wb, 0),
                NextValue(spi_cs_n, 1),
                NextValue(has_dummy, 0),
                NextValue(spi_do, 0),
                NextValue(spi_req, 0),
                NextValue(mac_count, 0),
                NextState("WAKEUP_PRE"),
                NextValue(new_cycle, 1),
                If(spi_mode, NextValue(bus_ack_r, 0)),
        )
        if spiread:
            mac.act("IDLE",
                If(spi_mode, # This machine stays in idle once spi_mode is dropped
                    NextValue(bus_ack_r, 0),
                    If((bus.cyc == 1) & (bus.stb == 1) & (bus.we == 0) & (bus.cti != 7), # read cycle requested, not end-of-burst
                        If( (rom_addr[2:] != bus.adr) & new_cycle,
                            NextValue(rom_addr, Cat(Signal(2, reset=0), bus.adr)),
                            NextValue(addr_updated, 1),
                            NextValue(spi_cs_n, 1), # raise CS in anticipation of a new address cycle
                            NextState("SPI_READ_32_CS"),
                        ).Elif( (rom_addr[2:] == bus.adr) | (~new_cycle & ((bus.cti == 2) | (bus.cti == 0)) ),
                            NextValue(mac_count, 3),  # get another beat of 4 bytes at the next address
                            NextState("SPI_READ_32")
                        ).Else(
                            NextValue(addr_updated, 0),
                            NextValue(spi_cs_n, 0),
                            NextState("SPI_READ_32"),
                            NextValue(mac_count, 3),  # prep the MAC state counter to count out 4 bytes
                        )
                    ).Elif(self.command.fields.wakeup,
                            NextValue(spi_cs_n, 1),
                            NextValue(self.command.storage, 0),  # clear all pending commands
                            NextState("WAKEUP_PRE"),
                    )
                )
            )
        else:
            mac.act("IDLE",
                If(spi_mode, # This machine stays in idle once spi_mode is dropped
                    If(self.command.fields.wakeup,
                        NextValue(spi_cs_n, 1),
                        NextValue(self.command.storage, 0),  # Clear all pending commands
                        NextState("WAKEUP_PRE"),
                    )
                )
            )

        #---------  wakup chip ------------------------------
        mac.act("WAKEUP_PRE",
            NextValue(spi_cs_n, 1), # Why isn't this sticking? i shouldn't have to put this here
            NextValue(mac_count, 4),
            NextState("WAKEUP_PRE_CS_WAIT")
        )
        mac.act("WAKEUP_PRE_CS_WAIT",
            NextValue(mac_count, mac_count-1),
            If(mac_count == 0,
               NextState("WAKEUP_WUP"),
               NextValue(spi_cs_n, 0)
            )
        )
        mac.act("WAKEUP_WUP",
            NextValue(mac_count, mac_count-1),
            If(mac_count == 0,
               NextValue(spi_cs_n, 0),
               NextValue(spi_do, 0xab),  # wakeup from deep sleep
               NextValue(spi_req, 1),
               NextState("WAKEUP_WUP_WAIT")
            )
        )
        mac.act("WAKEUP_WUP_WAIT",
            NextValue(spi_req, 0),
            If(spi_ack,
               NextValue(spi_cs_n, 1),   # raise CS
               NextValue(mac_count, 4),  # for >4 cycles per specsheet
               NextState("WAKEUP_CR2_WREN_1")
            )
        )

        #---------  WREN+CR2 - dummy cycles ------------------------------
        mac.act("WAKEUP_CR2_WREN_1",
            NextValue(mac_count, mac_count-1),
            If(mac_count == 0,
               NextValue(spi_cs_n, 0),
               NextValue(spi_do, 0x06),  # WREN to unlock CR2 writing
               NextValue(spi_req, 1),
               NextState("WAKEUP_CR2_WREN_1_WAIT")
            )
        )
        mac.act("WAKEUP_CR2_WREN_1_WAIT",
            NextValue(spi_req, 0),
            If(spi_ack,
               NextValue(spi_cs_n, 1),
               NextValue(mac_count, 4),
               NextState("WAKEUP_CR2_DUMMY_CMD")
            )
        )
        mac.act("WAKEUP_CR2_DUMMY_CMD",
            NextValue(mac_count, mac_count-1),
            If(mac_count == 0,
                NextValue(spi_cs_n, 0),
                NextValue(spi_do, 0x72), # CR2 command
                NextValue(spi_req, 1),
                NextValue(mac_count, 2),
                NextState("WAKEUP_CR2_DUMMY_ADRHI")
            )
        )
        mac.act("WAKEUP_CR2_DUMMY_ADRHI",
            NextValue(spi_do, 0x00), # We want to send 00_00_03_00
            If(spi_ack,
               NextValue(mac_count, mac_count -1)
            ),
            If(mac_count == 0,
               NextState("WAKEUP_CR2_DUMMY_ADRMID")
            )
        )
        mac.act("WAKEUP_CR2_DUMMY_ADRMID",
            NextValue(spi_do, 0x03),
            If(spi_ack,
                NextState("WAKEUP_CR2_DUMMY_ADRLO")
            )
        )
        mac.act("WAKEUP_CR2_DUMMY_ADRLO",
            NextValue(spi_do, 0x00),
            If(spi_ack,
                NextState("WAKEUP_CR2_DUMMY_DATA")
            )
        )
        mac.act("WAKEUP_CR2_DUMMY_DATA",
            NextValue(spi_do, 0x05), # 10 dummy cycles as required for 84MHz-104MHz operation
            If(spi_ack,
                NextState("WAKEUP_CR2_DUMMY_WAIT")
            ),
        )
        mac.act("WAKEUP_CR2_DUMMY_WAIT",
            NextValue(spi_req, 0),
            If(spi_ack,
                NextValue(spi_cs_n, 1),
                NextValue(mac_count, 4),
                NextState("WAKEUP_CR2_WREN_2")
            )
        )

        #---------  WREN+CR2 to DOPI mode ------------------------------
        mac.act("WAKEUP_CR2_WREN_2",
            NextValue(mac_count, mac_count-1),
            If(mac_count == 0,
                NextValue(spi_cs_n, 0),
                NextValue(spi_do, 0x06),  # WREN to unlock CR2 writing
                NextValue(spi_req, 1),
                NextState("WAKEUP_CR2_WREN_2_WAIT")
            )
        )
        mac.act("WAKEUP_CR2_WREN_2_WAIT",
            NextValue(spi_req, 0),
            If(spi_ack,
                NextValue(spi_cs_n, 1),
                NextValue(mac_count, 4),
                NextState("WAKEUP_CR2_DOPI_CMD")
            )
        )
        mac.act("WAKEUP_CR2_DOPI_CMD",
            NextValue(mac_count, mac_count-1),
            If(mac_count == 0,
                NextValue(spi_cs_n, 0),
                NextValue(spi_do, 0x72),  # CR2 command
                NextValue(spi_req, 1),
                NextValue(mac_count, 4),
                NextState("WAKEUP_CR2_DOPI_ADR")
            )
        )
        mac.act("WAKEUP_CR2_DOPI_ADR",  # send 0x00_00_00_00 as address
            NextValue(spi_do, 0x00),  # no need to raise CS or lower spi_req, this is back-to-back
            If(spi_ack,
               NextValue(mac_count, mac_count - 1)
            ),
            If(mac_count == 0,
               NextState("WAKEUP_CR2_DOPI_DATA")
            )
        ),
        mac.act("WAKEUP_CR2_DOPI_DATA",
            NextValue(spi_do, 2),    # enable DOPI mode
            If(spi_ack,
                NextState("WAKEUP_CR2_DOPI_WAIT")
            )
        )
        mac.act("WAKEUP_CR2_DOPI_WAIT", # trailing CS wait
            NextValue(spi_req, 0),
            If(spi_ack,
               NextValue(spi_cs_n, 1),
               NextValue(mac_count, 4),
               NextState("WAKEUP_CS_EXIT")
            )
        )
        mac.act("WAKEUP_CS_EXIT",
            NextValue(spi_mode, 0),  # now enter DOPI mode
            NextValue(mac_count, mac_count-1),
            If(mac_count == 0,
               NextState("IDLE")
            )
        )

        if spiread:
            #---------  SPI read machine ------------------------------
            mac.act("SPI_READ_32",
                If(addr_updated,
                    NextState("SPI_READ_32_CS"),
                    NextValue(has_dummy, 0),
                    NextValue(mac_count, 3),
                    NextValue(spi_cs_n, 1),
                    NextValue(spi_req, 0),
                ).Else(
                    If(mac_count > 0,
                        NextValue(has_dummy, 0),
                        NextValue(spi_req, 1),
                        NextState("SPI_READ_32_D")
                    ).Else(
                        NextValue(spi_req, 0),
                        If(spi_ack,
                            # Protect these in a spi_mode mux to prevent excess inference of logic to
                            # handle otherwise implicit dual-controller situation
                            If(spi_mode,
                                NextValue(bus.dat_r, Cat(d_to_wb[8:],spi_di)),
                                NextValue(bus_ack_r, 1),
                           ),
                           NextValue(rom_addr, rom_addr + 1),
                           NextState("IDLE")
                        )
                    )
                )
            )
            mac.act("SPI_READ_32_D",
                If(spi_ack,
                   # Shift in one byte at a time to d_to_wb(32)
                   NextValue(d_to_wb, Cat(d_to_wb[8:],spi_di,)),
                   NextValue(mac_count, mac_count - 1),
                   NextState("SPI_READ_32"),
                   NextValue(rom_addr, rom_addr + 1)
                )
            )
            mac.act("SPI_READ_32_CS",
                NextValue(mac_count, mac_count-1),
                If(mac_count == 0,
                   NextValue(spi_cs_n, 0),
                   NextState("SPI_READ_32_A0")
                )
            )
            mac.act("SPI_READ_32_A0",
                NextValue(spi_do, 0x0c), # 32-bit address write for "fast read" command
                NextValue(spi_req, 1),
                NextState("SPI_READ_32_A1")
            )
            mac.act("SPI_READ_32_A1",
                NextValue(spi_do, rom_addr[24:] & 0x7), # queue up MSB to send, leave req high; mask off unused high bits
                If(spi_ack,
                   NextState("SPI_READ_32_A2")
                )
            )
            mac.act("SPI_READ_32_A2",
                NextValue(spi_do, rom_addr[16:24]),
                If(spi_ack,
                   NextState("SPI_READ_32_A3")
                )
            )
            mac.act("SPI_READ_32_A3",
                NextValue(spi_do, rom_addr[8:16]),
                If(spi_ack,
                   NextState("SPI_READ_32_A4")
                )
            )
            mac.act("SPI_READ_32_A4",
                NextValue(spi_do, rom_addr[:8]),
                If(spi_ack,
                   NextState("SPI_READ_32_A5")
                )
            )
            mac.act("SPI_READ_32_A5",
                NextValue(spi_do, 0),
                If(spi_ack,
                   NextState("SPI_READ_32_DUMMY")
                )
            )
            mac.act("SPI_READ_32_DUMMY",
                NextValue(spi_req, 0),
                NextValue(addr_updated, 0),
                If(spi_ack,
                   NextState("SPI_READ_32"),
                   NextValue(mac_count, 3),  # Prep the MAC state counter to count out 4 bytes
                ).Else(
                    NextState("SPI_READ_32_DUMMY")
                )
            )

        # Handle ECS_n -----------------------------------------------------------------------------
        # treat ECS_N as an async signal -- just a "rough guide" of problems
        ecs_n = Signal()
        self.specials += MultiReg(pads.ecs_n, ecs_n)

        self.submodules.ev = EventManager()
        self.ev.ecc_error = EventSourceProcess(description="An ECC event has happened on the current block; triggered by falling edge of ECC_N")
        self.ev.finalize()
        self.comb += self.ev.ecc_error.trigger.eq(ecs_n)
        ecc_reported = Signal()
        ecs_n_delay  = Signal()
        ecs_pulse    = Signal()

        self.ecc_address = CSRStatus(fields=[
            CSRField("ecc_address", size=32, description="Address of the most recent ECC event")
        ])
        self.ecc_status = CSRStatus(fields=[
            CSRField("ecc_error",    size=1, description="Live status of the ECS_N bit (ECC error on current packet when low)"),
            CSRField("ecc_overflow", size=1, description="More than one ECS_N event has happened since th last time ecc_address was checked")
        ])

        self.comb += self.ecc_status.fields.ecc_error.eq(ecs_n)
        self.sync += [
            ecs_pulse.eq(ecs_n_delay & ~ecs_n), # falling edge -> positive pulse
            If(ecs_pulse,
               self.ecc_address.fields.ecc_address.eq(rom_addr),
               If(ecc_reported,
                  self.ecc_status.fields.ecc_overflow.eq(1)
               ).Else(
                   self.ecc_status.fields.ecc_overflow.eq(self.ecc_status.fields.ecc_overflow),
               )
            ).Else(
                self.ecc_address.fields.ecc_address.eq(self.ecc_address.fields.ecc_address),
                If(self.ecc_status.we,
                   self.ecc_status.fields.ecc_overflow.eq(0),
                ).Else(
                    self.ecc_status.fields.ecc_overflow.eq(self.ecc_status.fields.ecc_overflow),
                )
            )
        ]
        self.sync += [
            ecs_n_delay.eq(ecs_n),
            If(ecs_pulse,
               ecc_reported.eq(1)
            ).Elif(self.ecc_address.we,
                ecc_reported.eq(0)
            )
        ]

    def add_timing_constraints(self, platform, padgroup_name):
        # reminder to self: the {{ and }} overloading is because Python treats these as special in strings, so {{ -> { in actual constraint
        # NOTE: ECSn is deliberately not constrained -- it's more or less async (0-10ns delay on the signal, only meant to line up with "block" region

        # constrain DQS-to-DQ input DDR delays
        platform.add_platform_command("create_clock -name spidqs -period 10 [get_ports {}_dqs]".format(padgroup_name))
        platform.add_platform_command("set_input_delay -clock spidqs -max 0.6 [get_ports {{" + padgroup_name + "_dq[*]}}]")
        platform.add_platform_command("set_input_delay -clock spidqs -min 4.4 [get_ports {{" + padgroup_name + "_dq[*]}}]")
        platform.add_platform_command(
            "set_input_delay -clock spidqs -max 0.6 [get_ports {{" + padgroup_name + "_dq[*]}}] -clock_fall -add_delay")
        platform.add_platform_command(
            "set_input_delay -clock spidqs -min 4.4 [get_ports {{" + padgroup_name + "_dq[*]}}] -clock_fall -add_delay")

        # derive clock for SCLK - clock-forwarded from DDR see Xilinx answer 62488 use case #4
        platform.add_platform_command(
            "create_generated_clock -name spiclk_out -multiply_by 1 -source [get_pins {}/Q] [get_ports {}_sclk]".format(
                self.sclk_name, padgroup_name))

        # constrain CIPO SDR delay -- WARNING: -max is 'actually' 5.0ns, but design can't meet timing @ 5.0 tPD from SPIROM. There is some margin in the timing closure tho, so 4.5ns is probably going to work....
        platform.add_platform_command(
            "set_input_delay -clock [get_clocks spiclk_out] -clock_fall -max 4.5 [get_ports {}_dq[1]]".format(padgroup_name))
        platform.add_platform_command(
            "set_input_delay -clock [get_clocks spiclk_out] -clock_fall -min 1 [get_ports {}_dq[1]]".format(padgroup_name))
        # corresponding false path on CIPO DDR input when clocking SDR data
        platform.add_platform_command(
            "set_false_path -from [get_clocks spiclk_out] -to [get_pin {}/D ]".format(self.iddr_name + "1"))
        # corresponding false path on CIPO SDR input from DQS strobe, only if the cipo path is used
        if self.spiread:
            platform.add_platform_command(
                "set_false_path -from [get_clocks spidqs] -to [get_pin {}/D ]".format(self.cipo_name))

        # constrain CLK-to-DQ output DDR delays; copi uses the same rules
        platform.add_platform_command(
            "set_output_delay -clock [get_clocks spiclk_out] -max 1 [get_ports {{" + padgroup_name + "_dq[*]}}]")
        platform.add_platform_command(
            "set_output_delay -clock [get_clocks spiclk_out] -min -1 [get_ports {{" + padgroup_name + "_dq[*]}}]")
        platform.add_platform_command(
            "set_output_delay -clock [get_clocks spiclk_out] -max 1 [get_ports {{" + padgroup_name + "_dq[*]}}] -clock_fall -add_delay")
        platform.add_platform_command(
            "set_output_delay -clock [get_clocks spiclk_out] -min -1 [get_ports {{" + padgroup_name + "_dq[*]}}] -clock_fall -add_delay")
        # constrain CLK-to-CS output delay. NOTE: timings require one dummy cycle insertion between CS and SCLK (de)activations. Not possible to meet timing for DQ & single-cycle CS due to longer tS/tH reqs for CS
        platform.add_platform_command(
            "set_output_delay -clock [get_clocks spiclk_out] -min -1 [get_ports {}_cs_n]".format(padgroup_name))  # -3 in reality
        platform.add_platform_command(
            "set_output_delay -clock [get_clocks spiclk_out] -max 1 [get_ports {}_cs_n]".format(padgroup_name))  # 4.5 in reality
        # unconstrain OE path - we have like 10+ dummy cycles to turn the bus on wr->rd, and 2+ cycles to turn on end of read
        platform.add_platform_command("set_false_path -through [ get_pins {net}_reg/Q ]", net=self.dq.oe)
        platform.add_platform_command("set_false_path -through [ get_pins {net}_reg/Q ]",
            net=self.dq_copi.oe)



#
# This file is part of LiteX.
#
# Copyright (c) 2018 Dolu1990 <charles.papon.90@gmail.com>
# Copyright (c) 2018-2019 Florent Kermarrec <florent@enjoy-digital.fr>
# Copyright (c) 2018-2019 Sean Cross <sean@xobs.io>
# Copyright (c) 2019 Tim 'mithro' Ansell <me@mith.ro>
# Copyright (c) 2019 David Shah <dave@ds0.me>
# Copyright (c) 2019 Antmicro <www.antmicro.com>
# Copyright (c) 2019 Kurt Kiefer <kekiefer@gmail.com>
# SPDX-License-Identifier: BSD-2-Clause

import os

from migen import *

from litex import get_data_mod
from litex.soc.interconnect import wishbone
from litex.soc.interconnect.csr import *
from litex.soc.cores.cpu import CPU, CPU_GCC_TRIPLE_RISCV32

# Variants -----------------------------------------------------------------------------------------

CPU_VARIANTS = {
    "minimal":          "VexRiscv_Min",
    "minimal+debug":    "VexRiscv_MinDebug",
    "lite":             "VexRiscv_Lite",
    "lite+debug":       "VexRiscv_LiteDebug",
    "standard":         "VexRiscv",
    "standard+debug":   "VexRiscv_Debug",
    "imac":             "VexRiscv_IMAC",
    "imac+debug":       "VexRiscv_IMACDebug",
    "full":             "VexRiscv_Full",
    "full+cfu":         "VexRiscv_FullCfu",
    "full+debug":       "VexRiscv_FullDebug",
    "full+cfu+debug":   "VexRiscv_FullCfuDebug",
    "linux":            "VexRiscv_Linux",
    "linux+debug":      "VexRiscv_LinuxDebug",
    "linux+no-dsp":     "VexRiscv_LinuxNoDspFmax",
    "secure":           "VexRiscv_Secure",
    "secure+debug":     "VexRiscv_SecureDebug",
}

# GCC Flags ----------------------------------------------------------------------------------------

GCC_FLAGS = {
    #                               /-------- Base ISA
    #                               |/------- Hardware Multiply + Divide
    #                               ||/----- Atomics
    #                               |||/---- Compressed ISA
    #                               ||||/--- Single-Precision Floating-Point
    #                               |||||/-- Double-Precision Floating-Point
    #                               imacfd
    "minimal":          "-march=rv32i      -mabi=ilp32",
    "minimal+debug":    "-march=rv32i      -mabi=ilp32",
    "lite":             "-march=rv32im     -mabi=ilp32",
    "lite+debug":       "-march=rv32im     -mabi=ilp32",
    "standard":         "-march=rv32im     -mabi=ilp32",
    "standard+debug":   "-march=rv32im     -mabi=ilp32",
    "imac":             "-march=rv32imac   -mabi=ilp32",
    "imac+debug":       "-march=rv32imac   -mabi=ilp32",
    "full":             "-march=rv32im     -mabi=ilp32",
    "full+cfu":         "-march=rv32im     -mabi=ilp32",
    "full+debug":       "-march=rv32im     -mabi=ilp32",
    "full+cfu+debug":   "-march=rv32im     -mabi=ilp32",
    "linux":            "-march=rv32ima    -mabi=ilp32",
    "linux+debug":      "-march=rv32ima    -mabi=ilp32",
    "linux+no-dsp":     "-march=rv32ima    -mabi=ilp32",
    "secure":           "-march=rv32ima    -mabi=ilp32",
    "secure+debug":     "-march=rv32ima    -mabi=ilp32",
}

# VexRiscv Timer -----------------------------------------------------------------------------------

class VexRiscvTimer(Module, AutoCSR):
    def __init__(self):
        self._latch    = CSR()
        self._time     = CSRStatus(64)
        self._time_cmp = CSRStorage(64, reset=2**64-1)
        self.interrupt = Signal()

        # # #

        time = Signal(64)
        self.sync += time.eq(time + 1)
        self.sync += If(self._latch.re, self._time.status.eq(time))

        time_cmp = Signal(64, reset=2**64-1)
        self.sync += If(self._latch.re, time_cmp.eq(self._time_cmp.storage))

        self.comb += self.interrupt.eq(time >= time_cmp)

# VexRiscv -----------------------------------------------------------------------------------------

class VexRiscv(CPU, AutoCSR):
    name                 = "vexriscv"
    human_name           = "VexRiscv"
    variants             = CPU_VARIANTS
    data_width           = 32
    endianness           = "little"
    gcc_triple           = CPU_GCC_TRIPLE_RISCV32
    linker_output_format = "elf32-littleriscv"
    nop                  = "nop"
    io_regions           = {0x80000000: 0x80000000} # Origin, Length

    # Memory Mapping.
    @property
    def mem_map(self):
        return {
            "rom":            0x00000000,
            "sram":           0x10000000,
            "main_ram":       0x40000000,
            "csr":            0xf0000000,
            "vexriscv_debug": 0xf00f0000,
        }

    # GCC Flags.
    @property
    def gcc_flags(self):
        flags = GCC_FLAGS[self.variant]
        flags += " -D__vexriscv__"
        return flags

    def __init__(self, platform, variant="standard", with_timer=False):
        self.platform         = platform
        self.variant          = variant
        self.human_name       = CPU_VARIANTS.get(variant, "VexRiscv")
        self.external_variant = None
        self.reset            = Signal()
        self.interrupt        = Signal(32)
        self.ibus             = ibus = wishbone.Interface()
        self.dbus             = dbus = wishbone.Interface()
        self.periph_buses     = [ibus, dbus] # Peripheral buses (Connected to main SoC's bus).
        self.memory_buses     = []           # Memory buses (Connected directly to LiteDRAM).

        # # #

        # CPU Instance.
        self.cpu_params = dict(
            i_clk                    = ClockSignal("sys"),
            i_reset                  = ResetSignal("sys") | self.reset,

            i_externalInterruptArray = self.interrupt,
            i_timerInterrupt         = 0,
            i_softwareInterrupt      = 0,

            o_iBusWishbone_ADR      = ibus.adr,
            o_iBusWishbone_DAT_MOSI = ibus.dat_w,
            o_iBusWishbone_SEL      = ibus.sel,
            o_iBusWishbone_CYC      = ibus.cyc,
            o_iBusWishbone_STB      = ibus.stb,
            o_iBusWishbone_WE       = ibus.we,
            o_iBusWishbone_CTI      = ibus.cti,
            o_iBusWishbone_BTE      = ibus.bte,
            i_iBusWishbone_DAT_MISO = ibus.dat_r,
            i_iBusWishbone_ACK      = ibus.ack,
            i_iBusWishbone_ERR      = ibus.err,

            o_dBusWishbone_ADR      = dbus.adr,
            o_dBusWishbone_DAT_MOSI = dbus.dat_w,
            o_dBusWishbone_SEL      = dbus.sel,
            o_dBusWishbone_CYC      = dbus.cyc,
            o_dBusWishbone_STB      = dbus.stb,
            o_dBusWishbone_WE       = dbus.we,
            o_dBusWishbone_CTI      = dbus.cti,
            o_dBusWishbone_BTE      = dbus.bte,
            i_dBusWishbone_DAT_MISO = dbus.dat_r,
            i_dBusWishbone_ACK      = dbus.ack,
            i_dBusWishbone_ERR      = dbus.err
        )

        # Add Timer (Optional).
        if with_timer:
            self.add_timer()

        # Add Debug (Optional).
        if "debug" in variant:
            self.add_debug()

    def set_reset_address(self, reset_address):
        assert not hasattr(self, "reset_address")
        self.reset_address = reset_address
        self.cpu_params.update(i_externalResetVector=Signal(32, reset=reset_address))

    def add_timer(self):
        self.submodules.timer = VexRiscvTimer()
        self.cpu_params.update(i_timerInterrupt=self.timer.interrupt)

    def add_debug(self):
        debug_reset = Signal()

        ibus_err = Signal()
        dbus_err = Signal()

        self.i_cmd_valid           = Signal()
        self.i_cmd_payload_wr      = Signal()
        self.i_cmd_payload_address = Signal(8)
        self.i_cmd_payload_data    = Signal(32)
        self.o_cmd_ready           = Signal()
        self.o_rsp_data            = Signal(32)
        self.o_resetOut            = Signal()

        reset_debug_logic = Signal()

        self.transfer_complete     = Signal()
        self.transfer_in_progress  = Signal()
        self.transfer_wait_for_ack = Signal()

        self.debug_bus = wishbone.Interface()

        self.sync += self.debug_bus.dat_r.eq(self.o_rsp_data)
        self.sync += debug_reset.eq(reset_debug_logic | ResetSignal())

        self.sync += [
            # CYC is held high for the duration of the transfer.
            # STB is kept high when the transfer finishes (write)
            # or the master is waiting for data (read), and stays
            # there until ACK, ERR, or RTY are asserted.
            If((self.debug_bus.stb & self.debug_bus.cyc)
            & (~self.transfer_in_progress)
            & (~self.transfer_complete)
            & (~self.transfer_wait_for_ack),
                self.i_cmd_payload_data.eq(self.debug_bus.dat_w),
                self.i_cmd_payload_address.eq((self.debug_bus.adr[0:6] << 2) | 0),
                self.i_cmd_payload_wr.eq(self.debug_bus.we),
                self.i_cmd_valid.eq(1),
                self.transfer_in_progress.eq(1),
                self.transfer_complete.eq(0),
                self.debug_bus.ack.eq(0)
            ).Elif(self.transfer_in_progress,
                If(self.o_cmd_ready,
                    self.i_cmd_valid.eq(0),
                    self.i_cmd_payload_wr.eq(0),
                    self.transfer_complete.eq(1),
                    self.transfer_in_progress.eq(0)
                )
            ).Elif(self.transfer_complete,
                self.transfer_complete.eq(0),
                self.debug_bus.ack.eq(1),
                self.transfer_wait_for_ack.eq(1)
            ).Elif(self.transfer_wait_for_ack & ~(self.debug_bus.stb & self.debug_bus.cyc),
                self.transfer_wait_for_ack.eq(0),
                self.debug_bus.ack.eq(0)
            ),
            # Force a Wishbone error if transferring during a reset sequence.
            # Because o_resetOut is multiple cycles and i.stb/d.stb should
            # deassert one cycle after i_err/i_ack/d_err/d_ack are asserted,
            # this will give i_err and o_err enough time to be reset to 0
            # once the reset cycle finishes.
            If(self.o_resetOut,
                If(self.ibus.cyc & self.ibus.stb, ibus_err.eq(1)).Else(ibus_err.eq(0)),
                If(self.dbus.cyc & self.dbus.stb, dbus_err.eq(1)).Else(dbus_err.eq(0)),
                reset_debug_logic.eq(1))
            .Else(
                reset_debug_logic.eq(0)
            )
        ]

        self.cpu_params.update(
            i_reset = ResetSignal() | self.reset | debug_reset,
            i_iBusWishbone_ERR              = self.ibus.err | ibus_err,
            i_dBusWishbone_ERR              = self.dbus.err | dbus_err,
            i_debugReset                    = ResetSignal(),
            i_debug_bus_cmd_valid           = self.i_cmd_valid,
            i_debug_bus_cmd_payload_wr      = self.i_cmd_payload_wr,
            i_debug_bus_cmd_payload_address = self.i_cmd_payload_address,
            i_debug_bus_cmd_payload_data    = self.i_cmd_payload_data,
            o_debug_bus_cmd_ready           = self.o_cmd_ready,
            o_debug_bus_rsp_data            = self.o_rsp_data,
            o_debug_resetOut                = self.o_resetOut
        )

    def add_cfu(self, cfu_filename):
        # Check CFU presence.
        if not os.path.exists(cfu_filename):
            raise OSError(f"Unable to find VexRiscv CFU plugin {cfu_filename}.")

        # CFU Layout.
        cfu_bus_layout = [
            ("cmd", [
                ("valid", 1),
                ("ready", 1),
                ("payload", [
                    ("function_id", 10),
                    ("inputs_0", 32),
                    ("inputs_1", 32),
                ]),
            ]),
            ("rsp", [
                ("valid", 1),
                ("ready", 1),
                ("payload", [
                    ("response_ok", 1),
                    ("outputs_0", 32),
                ]),
            ]),
        ]

        # CFU Bus.
        self.cfu_bus = cfu_bus = Record(cfu_bus_layout)

        # Add CFU.
        self.specials += Instance("Cfu",
            i_cmd_valid                = cfu_bus.cmd.valid,
            o_cmd_ready                = cfu_bus.cmd.ready,
            i_cmd_payload_function_id  = cfu_bus.cmd.payload.function_id,
            i_cmd_payload_inputs_0     = cfu_bus.cmd.payload.inputs_0,
            i_cmd_payload_inputs_1     = cfu_bus.cmd.payload.inputs_1,
            o_rsp_valid                = cfu_bus.rsp.valid,
            i_rsp_ready                = cfu_bus.rsp.ready,
            o_rsp_payload_response_ok  = cfu_bus.rsp.payload.response_ok,
            o_rsp_payload_outputs_0    = cfu_bus.rsp.payload.outputs_0,
            i_clk                      = ClockSignal("sys"),
            i_reset                    = ResetSignal("sys"),
        )
        self.platform.add_source(cfu_filename)

        # Connect CFU to CPU.
        self.cpu_params.update(
            o_CfuPlugin_bus_cmd_valid                = cfu_bus.cmd.valid,
            i_CfuPlugin_bus_cmd_ready                = cfu_bus.cmd.ready,
            o_CfuPlugin_bus_cmd_payload_function_id  = cfu_bus.cmd.payload.function_id,
            o_CfuPlugin_bus_cmd_payload_inputs_0     = cfu_bus.cmd.payload.inputs_0,
            o_CfuPlugin_bus_cmd_payload_inputs_1     = cfu_bus.cmd.payload.inputs_1,
            i_CfuPlugin_bus_rsp_valid                = cfu_bus.rsp.valid,
            o_CfuPlugin_bus_rsp_ready                = cfu_bus.rsp.ready,
            i_CfuPlugin_bus_rsp_payload_response_ok  = cfu_bus.rsp.payload.response_ok,
            i_CfuPlugin_bus_rsp_payload_outputs_0    = cfu_bus.rsp.payload.outputs_0,
        )

    @staticmethod
    def add_sources(platform, variant="standard"):
        cpu_filename = CPU_VARIANTS[variant] + ".v"
        vdir = get_data_mod("cpu", "vexriscv").data_location
        platform.add_source(os.path.join(vdir, cpu_filename))

    def add_soc_components(self, soc, soc_region_cls):
        # Connect Debug interface to SoC.
        if "debug" in self.variant:
            soc.bus.add_slave("vexriscv_debug", self.debug_bus, region=
                soc_region_cls(
                    origin = soc.mem_map.get("vexriscv_debug"),
                    size   = 0x100,
                    cached = False
                )
            )

        # Pass I/D Caches info to software.
        base_variant = str(self.variant.split('+')[0])
        # DCACHE is present on all variants except minimal and lite.
        if not base_variant in ["minimal", "lite"]:
            soc.add_config("CPU_HAS_DCACHE")
        # ICACHE is present on all variants except minimal.
        if not base_variant in ["minimal"]:
            soc.add_config("CPU_HAS_ICACHE")

    def use_external_variant(self, variant_filename):
        self.external_variant = True
        self.platform.add_source(variant_filename)

    def do_finalize(self):
        assert hasattr(self, "reset_address")
        if not self.external_variant:
            self.add_sources(self.platform, self.variant)
        self.specials += Instance("VexRiscv", **self.cpu_params)



#
# This file is part of LiteX.
#
# This file is Copyright (c) 2018 Florent Kermarrec <florent@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

from migen import *


def reverse_bits(s):
    return s[::-1]


def reverse_bytes(s):
    n = (len(s) + 7)//8
    return Cat(*[s[i*8:min((i + 1)*8, len(s))]
        for i in reversed(range(n))])



####################################################################################################
#       DISCLAIMER: Provides retro-compatibility layer for stream_sim based designs.
#              Will soon no longer work, please don't use in new designs.
####################################################################################################

#
# This file is part of LiteX.
#
# Copyright (c) 2015-2018 Florent Kermarrec <florent@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

import random
import math
from copy import deepcopy

from migen import *

from litex.soc.interconnect import stream

__all__ = [
    "print_with_prefix",
    "seed_to_data",
    "split_bytes",
    "merge_bytes",
    "get_field_data",
    "comp",
    "check",
    "randn",
    "Packet",
    "PacketStreamer",
    "PacketLogger",
    "Randomizer"
]

def print_with_prefix(s, prefix=""):
    if not isinstance(s, str):
        s = s.__repr__()
    s = s.split("\n")
    for l in s:
        print(prefix + l)


def seed_to_data(seed, random=True):
    if random:
        return (seed * 0x31415979 + 1) & 0xffffffff
    else:
        return seed


def split_bytes(v, n, endianness="big"):
    r = []
    r_bytes = v.to_bytes(n, byteorder=endianness)
    for byte in r_bytes:
        r.append(int(byte))
    return r


def merge_bytes(b, endianness="big"):
    return int.from_bytes(bytes(b), endianness)


def get_field_data(field, datas):
    v = merge_bytes(datas[field.byte:field.byte+math.ceil(field.width/8)])
    return (v >> field.offset) & (2**field.width-1)


def comp(p1, p2):
    r = True
    for x, y in zip(p1, p2):
        if x != y:
            r = False
    return r


def check(p1, p2):
    p1 = deepcopy(p1)
    p2 = deepcopy(p2)
    if isinstance(p1, int):
        return 0, 1, int(p1 != p2)
    else:
        if len(p1) >= len(p2):
            ref, res = p1, p2
        else:
            ref, res = p2, p1
        shift = 0
        while((ref[0] != res[0]) and (len(res) > 1)):
            res.pop(0)
            shift += 1
        length = min(len(ref), len(res))
        errors = 0
        for i in range(length):
            if ref.pop(0) != res.pop(0):
                errors += 1
        return shift, length, errors


def randn(max_n):
    return random.randint(0, max_n-1)


class Packet(list):
    def __init__(self, init=[]):
        self.ongoing = False
        self.done = False
        for data in init:
            self.append(data)


class PacketStreamer(Module):
    def __init__(self, description, last_be=None, packet_cls=Packet):
        self.source = stream.Endpoint(description)
        self.last_be = last_be

        # # #

        self.packets = []
        self.packet = packet_cls()
        self.packet.done = True

    def send(self, packet):
        packet = deepcopy(packet)
        self.packets.append(packet)
        return packet

    def send_blocking(self, packet):
        packet = self.send(packet)
        while not packet.done:
            yield

    @passive
    def generator(self):
        while True:
            if len(self.packets) and self.packet.done:
                self.packet = self.packets.pop(0)
            if not self.packet.ongoing and not self.packet.done:
                yield self.source.valid.eq(1)
                yield self.source.data.eq(self.packet.pop(0))
                self.packet.ongoing = True
            elif (yield self.source.valid) and (yield self.source.ready):
                yield self.source.last.eq(len(self.packet) == 1)
                if self.last_be is not None:
                    yield self.source.last_be.eq(self.last_be & (len(self.packet) == 1))
                if len(self.packet):
                    yield self.source.valid.eq(1)
                    yield self.source.data.eq(self.packet.pop(0))
                else:
                    self.packet.done = True
                    yield self.source.valid.eq(0)
            yield


class PacketLogger(Module):
    def __init__(self, description, packet_cls=Packet):
        self.sink = stream.Endpoint(description)

        # # #

        self.packet_cls = packet_cls
        self.packet = packet_cls()
        self.first = True

    def receive(self, length=None):
        self.packet.done = False
        if length is None:
            while not self.packet.done:
                yield
        else:
            while length > len(self.packet):
                yield

    @passive
    def generator(self):
        while True:
            yield self.sink.ready.eq(1)
            if (yield self.sink.valid):
                if self.first:
                    self.packet = self.packet_cls()
                    self.first = False
                self.packet.append((yield self.sink.data))
                if (yield self.sink.last):
                    self.packet.done = True
                    self.first = True
            yield


class Randomizer(Module):
    def __init__(self, description, level=0):
        self.level = level

        self.sink = stream.Endpoint(description)
        self.source = stream.Endpoint(description)

        self.ce = Signal(reset=1)

        self.comb += \
            If(self.ce,
                self.sink.connect(self.source)
            ).Else(
                self.source.valid.eq(0),
                self.sink.ready.eq(0),
            )

    @passive
    def generator(self):
        while True:
            n = randn(100)
            if n < self.level:
                yield self.ce.eq(0)
            else:
                yield self.ce.eq(1)
            yield



import sys
import time

# Helpers ------------------------------------------------------------------------------------------

def colorer(s, color="bright"): # FIXME: Move colorer to litex.common?
    header  = {
        "bright": "\x1b[1m",
        "green":  "\x1b[32m",
        "cyan":   "\x1b[36m",
        "red":    "\x1b[31m",
        "yellow": "\x1b[33m",
        "underline": "\x1b[4m"}[color]
    trailer = "\x1b[0m"
    return header + str(s) + trailer

# Compat -------------------------------------------------------------------------------------------

def compat_notice(name, date, info=""):
    print("Compat: {name} is {deprecated} since {date} and will soon no longer work, please {update}. {info}".format(
        name       = colorer(name),
        deprecated = colorer("deprecated", color="red"),
        date       = colorer(date),
        update     = colorer("update", color="red"),
        info       = info,
    ), end="")
    # Annoy user to force update :)
    for i in range(10):
        time.sleep(0.2)
        print(".", end="")
        sys.stdout.flush()
    print("thanks :)")

def add_compat(location):
    # Integration.
    if location == "litex.soc.integration":
        class compat_soc_sdram:
            noticed = False
            def __getattr__(self, name):
                if not self.noticed:
                    compat_notice("SoCSDRAM", date="2020-03-24", info="Switch to SoCCore/add_sdram/soc_core_args instead.")
                    self.noticed = True
                from litex.compat import soc_sdram
                return getattr(soc_sdram, name)
        sys.modules["litex.soc.integration.soc_sdram"] = compat_soc_sdram()
    # Interconnect.
    if location == "litex.soc.interconnect":
        class compat_stream_sim:
            noticed = False
            def __getattr__(self, name):
                if not self.noticed:
                    compat_notice("stream_sim", date="2020-03-24", info="Code will not be replaced, copy it in your project to continue using it.")
                    self.noticed = True
                from litex.compat import stream_sim
                return getattr(stream_sim, name)
        sys.modules["litex.soc.interconnect.stream_sim"] = compat_stream_sim()

    # Cores.
    if location == "litex.soc.cores":
        class compat_up5kspram:
            noticed = False
            def __getattr__(self, name):
                if not self.noticed:
                    compat_notice("litex.soc.cores.up5kspram", date="2020-03-24", info="Switch to litex.soc.cores.ram.")
                    self.noticed = True
                from litex.soc.cores import ram
                return getattr(ram, name)
        from litex.soc.cores import ram
        sys.modules["litex.soc.cores.up5kspram"] = compat_up5kspram()



#
# This file is part of LiteX.
#
# Copyright (c) 2015-2019 Florent Kermarrec <florent@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

from migen import *
from migen.genlib.cdc import MultiReg

from litex.soc.interconnect.csr import *

# Pulse Width Modulation ---------------------------------------------------------------------------

class PWM(Module, AutoCSR):
    """Pulse Width Modulation

    Provides the minimal hardware to do Pulse Width Modulation.

    Pulse Width Modulation can be useful for various purposes: dim leds, regulate a fan, control
    an oscillator. Software can configure the PWM width and period and enable/disable it.
    """
    def __init__(self, pwm=None, clock_domain="sys", with_csr=True,
        default_enable = 0,
        default_width  = 0,
        default_period = 0):
        if pwm is None:
            self.pwm = pwm = Signal()
        self.enable = Signal(reset=default_enable)
        self.width  = Signal(32, reset=default_width)
        self.period = Signal(32, reset=default_period)

        # # #

        counter = Signal(32, reset_less=True)

        sync = getattr(self.sync, clock_domain)
        sync += [
            If(self.enable,
                counter.eq(counter + 1),
                If(counter < self.width,
                    pwm.eq(1)
                ).Else(
                    pwm.eq(0)
                ),
                If(counter >= (self.period - 1),
                    counter.eq(0)
                )
            ).Else(
                counter.eq(0),
                pwm.eq(0)
            )
        ]

        if with_csr:
            self.add_csr(clock_domain)

    def add_csr(self, clock_domain):
        self._enable = CSRStorage(description="""PWM Enable.\n
            Write ``1`` to enable PWM.""",
            reset = self.enable.reset)
        self._width  = CSRStorage(32, reset_less=True, description="""PWM Width.\n
            Defines the *Duty cycle* of the PWM. PWM is active high for *Width* ``{cd}_clk`` cycles and
            active low for *Period - Width* ``{cd}_clk`` cycles.""".format(cd=clock_domain),
            reset = self.width.reset)
        self._period = CSRStorage(32, reset_less=True, description="""PWM Period.\n
            Defines the *Period* of the PWM in ``{cd}_clk`` cycles.""".format(cd=clock_domain),
            reset = self.period.reset)

        n = 0 if clock_domain == "sys" else 2
        self.specials += [
            MultiReg(self._enable.storage, self.enable, n=n),
            MultiReg(self._width.storage,  self.width,  n=n),
            MultiReg(self._period.storage, self.period, n=n),
        ]



#
# This file is part of LiteX.
#
# Copyright (c) 2015-2019 Florent Kermarrec <florent@enjoy-digital.fr>
# Copyright (c) 2019 vytautasb <v.buitvydas@limemicro.com>
# SPDX-License-Identifier: BSD-2-Clause

from migen import *
from migen.fhdl.module import Module
from migen.fhdl.specials import Instance
from migen.genlib.resetsync import AsyncResetSynchronizer

from litex.build.io import *

# Common AsyncResetSynchronizer --------------------------------------------------------------------

class AlteraAsyncResetSynchronizerImpl(Module):
    def __init__(self, cd, async_reset):
        rst_meta = Signal()
        self.specials += [
            Instance("DFF",
                i_d    = 0,
                i_clk  = cd.clk,
                i_clrn = 1,
                i_prn  = ~async_reset,
                o_q    = rst_meta
            ),
            Instance("DFF",
                i_d    = rst_meta,
                i_clk  = cd.clk,
                i_clrn = 1,
                i_prn  = ~async_reset,
                o_q    = cd.rst
            )
        ]


class AlteraAsyncResetSynchronizer:
    @staticmethod
    def lower(dr):
        return AlteraAsyncResetSynchronizerImpl(dr.cd, dr.async_reset)

# Common DifferentialInput -------------------------------------------------------------------------

class AlteraDifferentialInputImpl(Module):
    def __init__(self, i_p, i_n, o):
        self.specials += [
            Instance("ALT_INBUF_DIFF",
                name   = "ibuf_diff",
                i_i    = i_p,
                i_ibar = i_n,
                o_o    = o
            )
        ]


class AlteraDifferentialInput:
    @staticmethod
    def lower(dr):
        return AlteraDifferentialInputImpl(dr.i_p, dr.i_n, dr.o)

# Common DifferentialOutput ------------------------------------------------------------------------

class AlteraDifferentialOutputImpl(Module):
    def __init__(self, i, o_p, o_n):
        self.specials += [
            Instance("ALT_OUTBUF_DIFF",
                name   = "obuf_diff",
                i_i    = i,
                o_o    = o_p,
                o_obar = o_n
            )
        ]


class AlteraDifferentialOutput:
    @staticmethod
    def lower(dr):
        return AlteraDifferentialOutputImpl(dr.i, dr.o_p, dr.o_n)

# Common DDROutput ---------------------------------------------------------------------------------

class AlteraDDROutputImpl(Module):
    def __init__(self, i1, i2, o, clk):
        self.specials += Instance("ALTDDIO_OUT",
            p_WIDTH    = 1,
            i_outclock = clk,
            i_datain_h = i1,
            i_datain_l = i2,
            o_dataout  = o,
        )


class AlteraDDROutput:
    @staticmethod
    def lower(dr):
        return AlteraDDROutputImpl(dr.i1, dr.i2, dr.o, dr.clk)

# Common DDRInput ----------------------------------------------------------------------------------

class AlteraDDRInputImpl(Module):
    def __init__(self, i, o1, o2, clk):
        self.specials += Instance("ALTDDIO_IN",
            p_WIDTH     = 1,
            i_inclock   = clk,
            i_datain    = i,
            o_dataout_h = o1,
            o_dataout_l = o2
        )

class AlteraDDRInput:
    @staticmethod
    def lower(dr):
        return AlteraDDRInputImpl(dr.i, dr.o1, dr.o2, dr.clk)

# Common SDROutput -------------------------------------------------------------------------------

class AlteraSDROutput:
    @staticmethod
    def lower(dr):
        return AlteraDDROutputImpl(dr.i, dr.i, dr.o, dr.clk)

# Common SDRInput --------------------------------------------------------------------------------

class AlteraSDRInput:
    @staticmethod
    def lower(dr):
        return AlteraDDRInputImpl(dr.i, dr.o, Signal(), dr.clk)

# Special Overrides --------------------------------------------------------------------------------

altera_special_overrides = {
    AsyncResetSynchronizer: AlteraAsyncResetSynchronizer,
    DifferentialInput:      AlteraDifferentialInput,
    DifferentialOutput:     AlteraDifferentialOutput,
    DDROutput:              AlteraDDROutput,
    DDRInput:               AlteraDDRInput,
    SDROutput:              AlteraSDROutput,
    SDRInput:               AlteraSDRInput,
}



#
# This file is part of LiteX.
#
# Copyright (c) 2020 Sean Cross <sean@xobs.io>
# SPDX-License-Identifier: BSD-2-Clause

import textwrap

def make_table(t):
    """Make a reStructured Text Table

    Returns
    -------

    A string containing a reStructured Text table.
    """
    column_widths = []

    table = "\n"
    if len(t) <= 0:
        return table

    # Figure out how wide to make each column
    for col in t[0]:
        column_widths.append(0)

    for row in t:
        for i, column in enumerate(row):
            column_widths[i] = max(column_widths[i], len(column))

    # Print out header
    header = t.pop(0)
    table += "+"
    for i, column in enumerate(header):
        table += "-" + "-"*column_widths[i]
        table += "-+"
    table += "\n"

    table += "|"
    for i, column in enumerate(header):
        table += " " + column.ljust(column_widths[i]) + " |"
    table += "\n"

    table += "+"
    for i, column in enumerate(header):
        table += "=" + "="*column_widths[i]
        table += "=+"
    table += "\n"

    for row in t:
        table += "|"
        for i, column in enumerate(row):
            table += " " + column.ljust(column_widths[i]) + " |"
        table += "\n"

        table += "+"
        for i, column in enumerate(row):
            table += "-" + "-"*column_widths[i]
            table += "-+"
        table += "\n"
    table += "\n"

    return table

def print_table(table, stream):
    """Print a reStructured Text table

    Arguments
    ---------

    table (:obj:`list` of :obj:`list`s): A list of rows in the table.
    Each row has several columns.  The first row is the table header.

    stream (:obj:`io`): Destination output file.
    """
    column_widths = []

    print("", file=stream)
    if len(table) <= 0:
        return

    # Figure out how wide to make each column
    for col in table[0]:
        column_widths.append(0)

    for row in table:
        for i, column in enumerate(row):
            column_widths[i] = max(column_widths[i], len(column))

    # Print out header
    header = table.pop(0)
    print("+", file=stream, end="")
    for i, column in enumerate(header):
        print("-" + "-"*column_widths[i], file=stream, end="")
        print("-+", file=stream, end="")
    print("", file=stream)

    print("|", file=stream, end="")
    for i, column in enumerate(header):
        print(" " + column.ljust(column_widths[i]) + " |", file=stream, end="")
    print("", file=stream)

    print("+", file=stream, end="")
    for i, column in enumerate(header):
        print("=" + "="*column_widths[i], file=stream, end="")
        print("=+", file=stream, end="")
    print("", file=stream)

    for row in table:
        print("|", file=stream, end="")
        for i, column in enumerate(row):
            print(" " + column.ljust(column_widths[i]) + " |", file=stream, end="")
        print("", file=stream)

        print("+", file=stream, end="")
        for i, column in enumerate(row):
            print("-" + "-"*column_widths[i], file=stream, end="")
            print("-+", file=stream, end="")
        print("", file=stream)
    print("", file=stream)

def pad_first_line_if_necessary(s):
    if not isinstance(s, str):
        return s
    lines = s.split("\n")

    # If there aren't at least two lines, don't do anything
    if len(lines) < 2:
        return s

    # If the first line is blank, don't do anything
    if lines[0].strip() == "":
        return s

    # If the pading on line 1 is greater than line 2, pad line 1
    # and return the result
    line_0_padding = len(lines[0]) - len(lines[0].lstrip(' '))
    line_1_padding = len(lines[1]) - len(lines[1].lstrip(' '))
    if (line_1_padding > 0) and (line_1_padding > line_0_padding):
        lines[0] = " " * (line_1_padding - line_0_padding) + lines[0]
        return "\n".join(lines)
    return s

def reflow(s, width=80):
    """Reflow the jagged text that gets generated as part
    of this Python comment.

    In this comment, the first line would be indented relative
    to the rest.  Additionally, the width of this block would
    be limited to the original text width.

    To reflow text, break it along \n\n, then dedent and reflow
    each line individually.

    Finally, append it to a new string to be returned.
    """
    if not isinstance(s, str):
        return s
    out = []
    s = pad_first_line_if_necessary(s)
    for piece in textwrap.dedent(s).split("\n\n"):
        trimmed_piece = textwrap.fill(textwrap.dedent(piece).strip(), width=width)
        out.append(trimmed_piece)
    return "\n\n".join(out)

def _reflow(s, width=80):
    return reflow(s, width)

def print_rst(stream, s, reflow=True):
    """Print a given string to the given stream.  Ensure it is reflowed."""
    print(_reflow(s), file=stream)
    print("", file=stream)



#
# This file is part of LiteX.
#
# Copyright (c) 2015 Sebastien Bourdeauducq <sb@m-labs.hk>
# Copyright (c) 2015-2020 Florent Kermarrec <florent@enjoy-digital.fr>
# Copyright (c) 2018 Tim 'mithro' Ansell <me@mith.ro>
# SPDX-License-Identifier: BSD-2-Clause

import math

from migen import *
from migen.util.misc import xdir
from migen.genlib import fifo
from migen.genlib.cdc import MultiReg, PulseSynchronizer

from litex.soc.interconnect.csr import *

# Endpoint -----------------------------------------------------------------------------------------

(DIR_SINK, DIR_SOURCE) = range(2)

def _make_m2s(layout):
    r = []
    for f in layout:
        if isinstance(f[1], (int, tuple)):
            r.append((f[0], f[1], DIR_M_TO_S))
        else:
            r.append((f[0], _make_m2s(f[1])))
    return r

def set_reset_less(field):
    if isinstance(field, Signal):
        field.reset_less = True
    elif isinstance(field, Record):
        for s, _ in field.iter_flat():
            s.reset_less = True

class EndpointDescription:
    def __init__(self, payload_layout=[], param_layout=[]):
        self.payload_layout = payload_layout
        self.param_layout   = param_layout

    def get_full_layout(self):
        reserved   = {"valid", "ready", "payload", "param", "first", "last", "description"}
        attributed = set()
        for f in self.payload_layout + self.param_layout:
            if f[0] in attributed:
                raise ValueError(f[0] + " already attributed in payload or param layout")
            if f[0] in reserved:
                raise ValueError(f[0] + " cannot be used in endpoint layout")
            attributed.add(f[0])

        full_layout = [
            ("valid",   1, DIR_M_TO_S),
            ("ready",   1, DIR_S_TO_M),
            ("first",   1, DIR_M_TO_S),
            ("last",    1, DIR_M_TO_S),
            ("payload", _make_m2s(self.payload_layout)),
            ("param",   _make_m2s(self.param_layout))
        ]
        return full_layout


class Endpoint(Record):
    def __init__(self, description_or_layout=[], name=None, **kwargs):
        if isinstance(description_or_layout, EndpointDescription):
            self.description = description_or_layout
        else:
            self.description = EndpointDescription(description_or_layout)
        Record.__init__(self, self.description.get_full_layout(), name, **kwargs)
        set_reset_less(self.first)
        set_reset_less(self.last)
        #set_reset_less(self.payload) # FIXME: cause issues with LiteSATA, understand why and uncomment.
        set_reset_less(self.param)

    def __getattr__(self, name):
        try:
            return getattr(object.__getattribute__(self, "payload"), name)
        except:
            return getattr(object.__getattribute__(self, "param"), name)

# Actor --------------------------------------------------------------------------------------------

def _rawbits_layout(l):
    if isinstance(l, int):
        return [("rawbits", l)]
    else:
        return l

def pack_layout(l, n):
    return [("chunk"+str(i), l) for i in range(n)]

def get_endpoints(obj, filt=Endpoint):
    if hasattr(obj, "get_endpoints") and callable(obj.get_endpoints):
        return obj.get_endpoints(filt)
    r = dict()
    for k, v in xdir(obj, True):
        if isinstance(v, filt):
            r[k] = v
    return r

def get_single_ep(obj, filt):
    eps = get_endpoints(obj, filt)
    if len(eps) != 1:
        raise ValueError("More than one endpoint")
    return list(eps.items())[0]


class BinaryActor(Module):
    def __init__(self, *args, **kwargs):
        self.build_binary_control(self.sink, self.source, *args, **kwargs)

    def build_binary_control(self, sink, source):
        raise NotImplementedError("Binary actor classes must overload build_binary_control_fragment")


class CombinatorialActor(BinaryActor):
    def build_binary_control(self, sink, source):
        self.comb += [
            source.valid.eq(sink.valid),
            source.first.eq(sink.first),
            source.last.eq(sink.last),
            sink.ready.eq(source.ready),
        ]


class PipelinedActor(BinaryActor):
    def __init__(self, latency):
        self.latency = latency
        self.pipe_ce = Signal()
        self.busy    = Signal()
        BinaryActor.__init__(self, latency)

    def build_binary_control(self, sink, source, latency):
        busy  = 0
        valid = sink.valid
        for i in range(latency):
            valid_n = Signal()
            self.sync += If(self.pipe_ce, valid_n.eq(valid))
            valid = valid_n
            busy = busy | valid

        self.comb += [
            self.pipe_ce.eq(source.ready | ~valid),
            sink.ready.eq(self.pipe_ce),
            source.valid.eq(valid),
            self.busy.eq(busy)
        ]
        first = sink.valid & sink.first
        last  = sink.valid & sink.last
        for i in range(latency):
            first_n = Signal(reset_less=True)
            last_n  = Signal(reset_less=True)
            self.sync += \
                If(self.pipe_ce,
                    first_n.eq(first),
                    last_n.eq(last)
                )
            first = first_n
            last  = last_n
        self.comb += [
            source.first.eq(first),
            source.last.eq(last)
        ]

# FIFO ---------------------------------------------------------------------------------------------

class _FIFOWrapper(Module):
    def __init__(self, fifo_class, layout, depth):
        self.sink   = sink   = Endpoint(layout)
        self.source = source = Endpoint(layout)

        # # #

        description = sink.description
        fifo_layout = [
            ("payload", description.payload_layout),
            ("param",   description.param_layout),
            ("first",   1),
            ("last",    1)
        ]

        self.submodules.fifo = fifo = fifo_class(layout_len(fifo_layout), depth)
        fifo_in  = Record(fifo_layout)
        fifo_out = Record(fifo_layout)
        self.comb += [
            fifo.din.eq(fifo_in.raw_bits()),
            fifo_out.raw_bits().eq(fifo.dout)
        ]

        self.comb += [
            sink.ready.eq(fifo.writable),
            fifo.we.eq(sink.valid),
            fifo_in.first.eq(sink.first),
            fifo_in.last.eq(sink.last),
            fifo_in.payload.eq(sink.payload),
            fifo_in.param.eq(sink.param),

            source.valid.eq(fifo.readable),
            source.first.eq(fifo_out.first),
            source.last.eq(fifo_out.last),
            source.payload.eq(fifo_out.payload),
            source.param.eq(fifo_out.param),
            fifo.re.eq(source.ready)
        ]


class SyncFIFO(_FIFOWrapper):
    def __init__(self, layout, depth, buffered=False):
        assert depth >= 0
        if depth >= 2:
            _FIFOWrapper.__init__(self,
                fifo_class = fifo.SyncFIFOBuffered if buffered else fifo.SyncFIFO,
                layout     = layout,
                depth      = depth)
            self.depth = self.fifo.depth
            self.level = self.fifo.level
        elif depth == 1:
            buf = Buffer(layout)
            self.submodules += buf
            self.sink   = buf.sink
            self.source = buf.source
            self.depth  = 1
            self.level  = Signal()
        elif depth == 0:
            self.sink   = Endpoint(layout)
            self.source = Endpoint(layout)
            self.comb += self.sink.connect(self.source)
            self.depth = 0
            self.level = Signal()


class AsyncFIFO(_FIFOWrapper):
    def __init__(self, layout, depth=None, buffered=False):
        depth = 4 if depth is None else depth
        assert depth >= 4
        _FIFOWrapper.__init__(self,
            fifo_class = fifo.AsyncFIFOBuffered if buffered else fifo.AsyncFIFO,
            layout     = layout,
            depth      = depth)

# ClockDomainCrossing ------------------------------------------------------------------------------

class ClockDomainCrossing(Module):
    def __init__(self, layout, cd_from="sys", cd_to="sys", depth=None):
        self.sink   = Endpoint(layout)
        self.source = Endpoint(layout)
        # # #

        if cd_from == cd_to:
            self.comb += self.sink.connect(self.source)
        else:
            cdc = AsyncFIFO(layout, depth)
            cdc = ClockDomainsRenamer({"write": cd_from, "read": cd_to})(cdc)
            self.submodules += cdc
            self.comb += self.sink.connect(cdc.sink)
            self.comb += cdc.source.connect(self.source)

# Mux/Demux ----------------------------------------------------------------------------------------

class Multiplexer(Module):
    def __init__(self, layout, n):
        self.source = Endpoint(layout)
        sinks = []
        for i in range(n):
            sink = Endpoint(layout)
            setattr(self, "sink"+str(i), sink)
            sinks.append(sink)
        self.sel = Signal(max=max(n, 2))

        # # #

        cases = {}
        for i, sink in enumerate(sinks):
            cases[i] = sink.connect(self.source)
        self.comb += Case(self.sel, cases)


class Demultiplexer(Module):
    def __init__(self, layout, n):
        self.sink = Endpoint(layout)
        sources = []
        for i in range(n):
            source = Endpoint(layout)
            setattr(self, "source"+str(i), source)
            sources.append(source)
        self.sel = Signal(max=max(n, 2))

        # # #

        cases = {}
        for i, source in enumerate(sources):
            cases[i] = self.sink.connect(source)
        self.comb += Case(self.sel, cases)


# Gate ---------------------------------------------------------------------------------------------

class Gate(Module):
    def __init__(self, layout, sink_ready_when_disabled=False):
        self.sink   = Endpoint(layout)
        self.source = Endpoint(layout)
        self.enable = Signal()

        # # #

        self.comb += [
            If(self.enable,
                self.sink.connect(self.source)
            ).Else(
                self.sink.ready.eq(int(sink_ready_when_disabled))
            )
        ]

# Converter ----------------------------------------------------------------------------------------

class _UpConverter(Module):
    def __init__(self, nbits_from, nbits_to, ratio, reverse):
        self.sink   = sink   = Endpoint([("data", nbits_from)])
        self.source = source = Endpoint([("data", nbits_to), ("valid_token_count", bits_for(ratio))])
        self.latency = 1

        # # #

        # Control path
        demux      = Signal(max=ratio)
        load_part  = Signal()
        strobe_all = Signal()
        self.comb += [
            sink.ready.eq(~strobe_all | source.ready),
            source.valid.eq(strobe_all),
            load_part.eq(sink.valid & sink.ready)
        ]

        demux_last = ((demux == (ratio - 1)) | sink.last)

        self.sync += [
            If(source.ready, strobe_all.eq(0)),
            If(load_part,
                If(demux_last,
                    demux.eq(0),
                    strobe_all.eq(1)
                ).Else(
                    demux.eq(demux + 1)
                )
            ),
            If(source.valid & source.ready,
                If(sink.valid & sink.ready,
                    source.first.eq(sink.first),
                    source.last.eq(sink.last)
                ).Else(
                    source.first.eq(0),
                    source.last.eq(0)
                )
            ).Elif(sink.valid & sink.ready,
                source.first.eq(sink.first | source.first),
                source.last.eq(sink.last | source.last)
            )
        ]

        # Data path
        cases = {}
        for i in range(ratio):
            n = ratio-i-1 if reverse else i
            cases[i] = source.data[n*nbits_from:(n+1)*nbits_from].eq(sink.data)
        self.sync += If(load_part, Case(demux, cases))

        # Valid token count
        self.sync += If(load_part, source.valid_token_count.eq(demux + 1))


class _DownConverter(Module):
    def __init__(self, nbits_from, nbits_to, ratio, reverse):
        self.sink   = sink   = Endpoint([("data", nbits_from)])
        self.source = source = Endpoint([("data", nbits_to), ("valid_token_count", 1)])
        self.latency = 0

        # # #

        # Control path
        mux   = Signal(max=ratio)
        first = Signal()
        last  = Signal()
        self.comb += [
            first.eq(mux == 0),
            last.eq(mux == (ratio-1)),
            source.valid.eq(sink.valid),
            source.first.eq(sink.first & first),
            source.last.eq(sink.last & last),
            sink.ready.eq(last & source.ready)
        ]
        self.sync += \
            If(source.valid & source.ready,
                If(last,
                    mux.eq(0)
                ).Else(
                    mux.eq(mux + 1)
                )
            )

        # Data path
        cases = {}
        for i in range(ratio):
            n = ratio-i-1 if reverse else i
            cases[i] = source.data.eq(sink.data[n*nbits_to:(n+1)*nbits_to])
        self.comb += Case(mux, cases).makedefault()

        # Valid token count
        self.comb += source.valid_token_count.eq(last)


class _IdentityConverter(Module):
    def __init__(self, nbits_from, nbits_to, ratio, reverse):
        self.sink   = sink   = Endpoint([("data", nbits_from)])
        self.source = source = Endpoint([("data", nbits_to), ("valid_token_count", 1)])
        self.latency = 0

        # # #

        self.comb += [
            sink.connect(source),
            source.valid_token_count.eq(1)
        ]


def _get_converter_ratio(nbits_from, nbits_to):
    if nbits_from > nbits_to:
        converter_cls = _DownConverter
        if nbits_from % nbits_to:
            raise ValueError("Ratio must be an int")
        ratio = nbits_from//nbits_to
    elif nbits_from < nbits_to:
        converter_cls = _UpConverter
        if nbits_to % nbits_from:
            raise ValueError("Ratio must be an int")
        ratio = nbits_to//nbits_from
    else:
        converter_cls = _IdentityConverter
        ratio = 1
    return converter_cls, ratio


class Converter(Module):
    def __init__(self, nbits_from, nbits_to,
        reverse                  = False,
        report_valid_token_count = False):
        self.cls, self.ratio = _get_converter_ratio(nbits_from, nbits_to)

        # # #

        converter = self.cls(nbits_from, nbits_to, self.ratio, reverse)
        self.submodules += converter
        self.latency = converter.latency

        self.sink = converter.sink
        if report_valid_token_count:
            self.source = converter.source
        else:
            self.source = Endpoint([("data", nbits_to)])
            self.comb += converter.source.connect(self.source, omit=set(["valid_token_count"]))


class StrideConverter(Module):
    def __init__(self, description_from, description_to, reverse=False):
        self.sink   = sink   = Endpoint(description_from)
        self.source = source = Endpoint(description_to)

        # # #

        nbits_from = len(sink.payload.raw_bits())
        nbits_to   = len(source.payload.raw_bits())

        converter = Converter(nbits_from, nbits_to, reverse)
        self.submodules += converter

        # Cast sink to converter.sink (user fields --> raw bits)
        self.comb += [
            converter.sink.valid.eq(sink.valid),
            converter.sink.first.eq(sink.first),
            converter.sink.last.eq(sink.last),
            sink.ready.eq(converter.sink.ready)
        ]
        if converter.cls == _DownConverter:
            ratio = converter.ratio
            for i in range(ratio):
                j = 0
                for name, width in source.description.payload_layout:
                    src = getattr(sink, name)[i*width:(i+1)*width]
                    dst = converter.sink.data[i*nbits_to+j:i*nbits_to+j+width]
                    self.comb += dst.eq(src)
                    j += width
        else:
            self.comb += converter.sink.data.eq(sink.payload.raw_bits())


        # Cast converter.source to source (raw bits --> user fields)
        self.comb += [
            source.valid.eq(converter.source.valid),
            source.first.eq(converter.source.first),
            source.last.eq(converter.source.last),
            converter.source.ready.eq(source.ready)
        ]
        if converter.cls == _UpConverter:
            ratio = converter.ratio
            for i in range(ratio):
                j = 0
                for name, width in sink.description.payload_layout:
                    src = converter.source.data[i*nbits_from+j:i*nbits_from+j+width]
                    dst = getattr(source, name)[i*width:(i+1)*width]
                    self.comb += dst.eq(src)
                    j += width
        else:
            self.comb += source.payload.raw_bits().eq(converter.source.data)

        # Connect params
        if converter.latency == 0:
            self.comb += source.param.eq(sink.param)
        elif converter.latency == 1:
            self.sync += source.param.eq(sink.param)
        else:
            raise ValueError

# Gearbox ------------------------------------------------------------------------------------------

def lcm(a, b):
    return (a*b)//math.gcd(a, b)


def inc_mod(s, m):
    return [s.eq(s + 1), If(s == (m -1), s.eq(0))]


class Gearbox(Module):
    def __init__(self, i_dw, o_dw, msb_first=True):
        self.sink   = sink   = Endpoint([("data", i_dw)])
        self.source = source = Endpoint([("data", o_dw)])

        # # #

        io_lcm = lcm(i_dw, o_dw)
        if (io_lcm//i_dw) < 2:
            io_lcm = io_lcm * 2
        if (io_lcm//o_dw) < 2:
            io_lcm = io_lcm * 2

        # Control path

        level   = Signal(max=io_lcm)
        i_inc   = Signal()
        i_count = Signal(max=io_lcm//i_dw)
        o_inc   = Signal()
        o_count = Signal(max=io_lcm//o_dw)

        self.comb += [
            sink.ready.eq(level < (io_lcm - i_dw)),
            source.valid.eq(level >= o_dw),
        ]
        self.comb += [
            i_inc.eq(sink.valid & sink.ready),
            o_inc.eq(source.valid & source.ready)
        ]
        self.sync += [
            If(i_inc, *inc_mod(i_count, io_lcm//i_dw)),
            If(o_inc, *inc_mod(o_count, io_lcm//o_dw)),
            If(i_inc & ~o_inc, level.eq(level + i_dw)),
            If(~i_inc & o_inc, level.eq(level - o_dw)),
            If(i_inc & o_inc, level.eq(level + i_dw - o_dw)),
        ]

        # Data path

        shift_register = Signal(io_lcm, reset_less=True)

        i_cases = {}
        i_data  = Signal(i_dw)
        if msb_first:
            self.comb += i_data.eq(sink.data)
        else:
            self.comb += i_data.eq(sink.data[::-1])
        for i in range(io_lcm//i_dw):
            i_cases[i] = shift_register[io_lcm - i_dw*(i+1):io_lcm - i_dw*i].eq(i_data)
        self.sync += If(sink.valid & sink.ready, Case(i_count, i_cases))

        o_cases = {}
        o_data  = Signal(o_dw)
        for i in range(io_lcm//o_dw):
            o_cases[i] = o_data.eq(shift_register[io_lcm - o_dw*(i+1):io_lcm - o_dw*i])
        self.comb += Case(o_count, o_cases)
        if msb_first:
            self.comb += source.data.eq(o_data)
        else:
            self.comb += source.data.eq(o_data[::-1])

# Shifter ------------------------------------------------------------------------------------------

class Shifter(PipelinedActor):
    def __init__(self, dw, shift=None):
        self.shift  = Signal(max=dw) if shift is None else shift
        self.sink   = sink   = Endpoint([("data", dw)])
        self.source = source = Endpoint([("data", dw)])
        PipelinedActor.__init__(self, latency=2)

        # # #

        # Accumulate current/last sink.data.
        r = Signal(2*dw)
        self.sync += If(self.pipe_ce,
            r[:dw].eq(r[dw:]),
            r[dw:].eq(sink.data)
        )

        # Select output data based on shift.
        cases = {}
        for i in range(dw):
            cases[i] = self.source.data.eq(r[i:dw+i])
        self.comb += Case(self.shift, cases)

# Monitor ------------------------------------------------------------------------------------------

class Monitor(Module, AutoCSR):
    def __init__(self, endpoint, count_width=32, clock_domain="sys",
        with_tokens     = False,
        with_overflows  = False,
        with_underflows = False):

        self.reset = CSR()
        self.latch = CSR()
        if with_tokens:
            self.tokens = CSRStatus(count_width)
        if with_overflows:
            self.overflows = CSRStatus(count_width)
        if with_underflows:
            self.underflows = CSRStatus(count_width)

        # # #

        reset = Signal()
        latch = Signal()
        if clock_domain == "sys":
            self.comb += reset.eq(self.reset.re)
            self.comb += latch.eq(self.latch.re)
        else:
            reset_ps = PulseSynchronizer("sys", clock_domain)
            latch_ps = PulseSynchronizer("sys", clock_domain)
            self.submodules += reset_ps, latch_ps
            self.comb += reset_ps.i.eq(self.reset.re)
            self.comb += reset.eq(reset_ps.o)
            self.comb += latch_ps.i.eq(self.latch.re)
            self.comb += latch.eq(latch_ps.o)

        # Generic Monitor Counter ------------------------------------------------------------------
        class MonitorCounter(Module):
            def __init__(self, reset, latch, enable, count):
                _count         = Signal.like(count)
                _count_latched = Signal.like(count)
                _sync = getattr(self.sync, clock_domain)
                _sync += [
                    If(reset,
                        _count.eq(0),
                        _count_latched.eq(0),
                    ).Elif(enable,
                        If(_count != (2**len(count)-1),
                            _count.eq(_count + 1)
                        )
                    ),
                    If(latch,
                        _count_latched.eq(_count)
                    )
                ]
                self.specials += MultiReg(_count_latched, count)

        # Tokens Count -----------------------------------------------------------------------------
        if with_tokens:
            token_counter = MonitorCounter(reset, latch, endpoint.valid & endpoint.ready, self.tokens.status)
            self.submodules += token_counter

        # Overflows Count (only useful when endpoint is expected to always be ready) ---------------
        if with_overflows:
            overflow_counter = MonitorCounter(reset, latch, endpoint.valid & ~endpoint.ready, self.overflows.status)
            self.submodules += overflow_counter

        # Underflows Count (only useful when endpoint is expected to always be valid) --------------
        if with_underflows:
            underflow_counter = MonitorCounter(reset, latch, ~endpoint.valid & endpoint.ready, self.underflows.status)
            self.submodules += underflow_counter

# Pipe ---------------------------------------------------------------------------------------------

class PipeValid(Module):
    """Pipe valid/payload to cut timing path"""
    def __init__(self, layout):
        self.sink   = sink   = Endpoint(layout)
        self.source = source = Endpoint(layout)

        # # #

        # Pipe when source is not valid or is ready.
        self.sync += [
            If(~source.valid | source.ready,
                source.valid.eq(sink.valid),
                source.first.eq(sink.first),
                source.last.eq(sink.last),
                source.payload.eq(sink.payload),
                source.param.eq(sink.param),
            )
        ]
        self.comb += sink.ready.eq(~source.valid | source.ready)


class PipeReady(Module):
    """Pipe ready to cut timing path"""
    def __init__(self, layout):
        self.sink   = sink   = Endpoint(layout)
        self.source = source = Endpoint(layout)

        # # #

        valid  = Signal()
        sink_d = Endpoint(layout)

        self.sync += [
            If(sink.valid & ~source.ready,
                valid.eq(1)
            ).Elif(source.ready,
                valid.eq(0)
            ),
            If(~source.ready & ~valid,
                sink_d.eq(sink)
            )
        ]
        self.comb += [
            sink.ready.eq(~valid),
            If(valid,
                sink_d.connect(source, omit={"ready"})
            ).Else(
                sink.connect(source, omit={"ready"})
            )
        ]

# Buffer -------------------------------------------------------------------------------------------

class Buffer(PipeValid): pass # FIXME: Replace Buffer with PipeValid in codebase?

# Cast ---------------------------------------------------------------------------------------------

class Cast(CombinatorialActor):
    def __init__(self, layout_from, layout_to, reverse_from=False, reverse_to=False):
        self.sink   = Endpoint(_rawbits_layout(layout_from))
        self.source = Endpoint(_rawbits_layout(layout_to))
        CombinatorialActor.__init__(self)

        # # #

        sigs_from = self.sink.payload.flatten()
        if reverse_from:
            sigs_from = list(reversed(sigs_from))
        sigs_to = self.source.payload.flatten()
        if reverse_to:
            sigs_to = list(reversed(sigs_to))
        if sum(len(s) for s in sigs_from) != sum(len(s) for s in sigs_to):
            raise TypeError
        self.comb += Cat(*sigs_to).eq(Cat(*sigs_from))

# Unpack/Pack --------------------------------------------------------------------------------------

class Unpack(Module):
    def __init__(self, n, layout_to, reverse=False):
        self.source = source = Endpoint(layout_to)
        description_from = Endpoint(layout_to).description
        description_from.payload_layout = pack_layout(description_from.payload_layout, n)
        self.sink = sink = Endpoint(description_from)

        # # #

        mux   = Signal(max=n)
        first = Signal()
        last  = Signal()
        self.comb += [
            first.eq(mux == 0),
            last.eq(mux == (n-1)),
            source.valid.eq(sink.valid),
            sink.ready.eq(last & source.ready)
        ]
        self.sync += [
            If(source.valid & source.ready,
                If(last,
                    mux.eq(0)
                ).Else(
                    mux.eq(mux + 1)
                )
            )
        ]
        cases = {}
        for i in range(n):
            chunk = n-i-1 if reverse else i
            cases[i] = [source.payload.raw_bits().eq(getattr(sink.payload, "chunk"+str(chunk)).raw_bits())]
        self.comb += Case(mux, cases).makedefault()

        for f in description_from.param_layout:
            src = getattr(self.sink, f[0])
            dst = getattr(self.source, f[0])
            self.comb += dst.eq(src)

        self.comb += [
            source.first.eq(sink.first & first),
            source.last.eq(sink.last & last)
        ]


class Pack(Module):
    def __init__(self, layout_from, n, reverse=False):
        self.sink = sink = Endpoint(layout_from)
        description_to = Endpoint(layout_from).description
        description_to.payload_layout = pack_layout(description_to.payload_layout, n)
        self.source = source = Endpoint(description_to)

        # # #

        demux = Signal(max=n)

        load_part  = Signal()
        strobe_all = Signal()
        cases = {}
        for i in range(n):
            chunk = n-i-1 if reverse else i
            cases[i] = [getattr(source.payload, "chunk"+str(chunk)).raw_bits().eq(sink.payload.raw_bits())]
        self.comb += [
            sink.ready.eq(~strobe_all | source.ready),
            source.valid.eq(strobe_all),
            load_part.eq(sink.valid & sink.ready)
        ]

        for f in description_to.param_layout:
            src = getattr(self.sink, f[0])
            dst = getattr(self.source, f[0])
            self.sync += If(load_part, dst.eq(src))

        demux_last = ((demux == (n - 1)) | sink.last)

        self.sync += [
            If(source.ready, strobe_all.eq(0)),
            If(load_part,
                Case(demux, cases),
                If(demux_last,
                    demux.eq(0),
                    strobe_all.eq(1)
                ).Else(
                    demux.eq(demux + 1)
                )
            ),
            If(source.valid & source.ready,
                source.first.eq(sink.first),
                source.last.eq(sink.last),
            ).Elif(sink.valid & sink.ready,
                source.first.eq(sink.first | source.first),
                source.last.eq(sink.last | source.last)
            )
        ]

# Pipeline -----------------------------------------------------------------------------------------

class Pipeline(Module):
    def __init__(self, *modules):
        n = len(modules)
        m = modules[0]
        # Expose sink of first module if available.
        if hasattr(m, "sink"):
            self.sink = m.sink
        # Iterate on Modules/Endpoints.
        for i in range(1, n):
            m_n = modules[i]
            # If m is an Endpoint, use it as Source, else use Module.source.
            source = m if isinstance(m, Endpoint) else m.source
            # If m_n is an Endpoint, use it as Sink, else use Module.sink.
            sink = m_n if isinstance(m_n, Endpoint) else m_n.sink
            # Connect Source to Sink (when m is not m_n).
            if m is not m_n:
                self.comb += source.connect(sink)
            # Update m.
            m = m_n
        # Expose source of last module if available.
        if hasattr(m, "source"):
            self.source = m.source

# BufferizeEndpoints -------------------------------------------------------------------------------

# Add buffers on Endpoints (can be used to improve timings)
class BufferizeEndpoints(ModuleTransformer):
    def __init__(self, endpoint_dict):
        self.endpoint_dict = endpoint_dict

    def transform_instance(self, submodule):
        for name, direction in self.endpoint_dict.items():
            endpoint = getattr(submodule, name)
            # add buffer on sinks
            if direction == DIR_SINK:
                buf = Buffer(endpoint.description)
                submodule.submodules += buf
                setattr(submodule, name, buf.sink)
                submodule.comb += buf.source.connect(endpoint)
            # add buffer on sources
            elif direction == DIR_SOURCE:
                buf = Buffer(endpoint.description)
                submodule.submodules += buf
                submodule.comb += endpoint.connect(buf.sink)
                setattr(submodule, name, buf.source)
            else:
                raise ValueError



#
# This file is part of LiteX.
#
# Copyright (c) 2019 Florent Kermarrec <florent@enjoy-digital.fr>
# Copyright (c) 2019 Benjamin Herrenschmidt <benh@ozlabs.org>
# Copyright (c) 2020 Raptor Engineering <sales@raptorengineering.com>
# SPDX-License-Identifier: BSD-2-Clause

import os

from migen import *

from litex import get_data_mod
from litex.soc.interconnect import wishbone
from litex.soc.interconnect.csr import *
from litex.gen.common import reverse_bytes
from litex.soc.cores.cpu import CPU

class Open(Signal): pass

# Variants -----------------------------------------------------------------------------------------

CPU_VARIANTS = ["standard", "standard+ghdl", "standard+irq", "standard+ghdl+irq"]

# Microwatt ----------------------------------------------------------------------------------------

class Microwatt(CPU):
    name                 = "microwatt"
    human_name           = "Microwatt"
    variants             = CPU_VARIANTS
    data_width           = 64
    endianness           = "little"
    gcc_triple           = ("powerpc64le-linux", "powerpc64le-linux-gnu")
    linker_output_format = "elf64-powerpcle"
    nop                  = "nop"
    io_regions           = {0xc0000000: 0x10000000} # Origin, Length.

    # Memory Mapping.
    @property
    def mem_map(self):
        return {
            # Keep the lower 128MBs for SoC IOs auto-allocation.
            "csr":      0xc8000000,
            "xicsicp":  0xcbff0000,
            "xicsics":  0xcbff1000,
        }

    # GCC Flags.
    @property
    def gcc_flags(self):
        flags  = "-m64 "
        flags += "-mabi=elfv2 "
        flags += "-msoft-float "
        flags += "-mno-string "
        flags += "-mno-multiple "
        flags += "-mno-vsx "
        flags += "-mno-altivec "
        flags += "-mlittle-endian "
        flags += "-mstrict-align "
        flags += "-fno-stack-protector "
        flags += "-D__microwatt__ "
        return flags

    def __init__(self, platform, variant="standard"):
        self.platform     = platform
        self.variant      = variant
        self.reset        = Signal()
        self.ibus         = ibus = wishbone.Interface(data_width=64, adr_width=29)
        self.dbus         = dbus = wishbone.Interface(data_width=64, adr_width=29)
        self.periph_buses = [ibus, dbus] # Peripheral buses (Connected to main SoC's bus).
        self.memory_buses = []           # Memory buses (Connected directly to LiteDRAM).
        if "irq" in variant:
            self.interrupt = Signal(16)
        self.core_ext_irq = Signal()

        # # #

        self.cpu_params = dict(
            # Clk / Rst.
            i_clk = ClockSignal("sys"),
            i_rst = ResetSignal("sys") | self.reset,

            # IBus.
            i_wishbone_insn_dat_r = ibus.dat_r,
            i_wishbone_insn_ack   = ibus.ack,
            i_wishbone_insn_stall = ibus.cyc & ~ibus.ack, # No burst support

            o_wishbone_insn_adr   = Cat(Signal(3), ibus.adr),
            o_wishbone_insn_dat_w = ibus.dat_w,
            o_wishbone_insn_cyc   = ibus.cyc,
            o_wishbone_insn_stb   = ibus.stb,
            o_wishbone_insn_sel   = ibus.sel,
            o_wishbone_insn_we    = ibus.we,

            # DBus.
            i_wishbone_data_dat_r = dbus.dat_r,
            i_wishbone_data_ack   = dbus.ack,
            i_wishbone_data_stall = dbus.cyc & ~dbus.ack, # No burst support

            o_wishbone_data_adr   = Cat(Signal(3), dbus.adr),
            o_wishbone_data_dat_w = dbus.dat_w,
            o_wishbone_data_cyc   = dbus.cyc,
            o_wishbone_data_stb   = dbus.stb,
            o_wishbone_data_sel   = dbus.sel,
            o_wishbone_data_we    = dbus.we,

            # Debug.
            i_dmi_addr = 0,
            i_dmi_din  = 0,
            o_dmi_dout = Open(),
            i_dmi_req  = 0,
            i_dmi_wr   = 0,
            o_dmi_ack  = Open(),

            # IRQ.
            i_core_ext_irq = self.core_ext_irq,
        )

        # Add VHDL sources.
        self.add_sources(platform, use_ghdl_yosys_plugin="ghdl" in self.variant)

    def set_reset_address(self, reset_address):
        assert not hasattr(self, "reset_address")
        self.reset_address = reset_address
        assert reset_address == 0x00000000

    def add_soc_components(self, soc, soc_region_cls):
        if "irq" in self.variant:
            self.submodules.xics = XICSSlave(
                platform     = self.platform,
                variant      = self.variant,
                core_irq_out = self.core_ext_irq,
                int_level_in = self.interrupt,
            )
            xicsicp_region = soc_region_cls(origin=soc.mem_map.get("xicsicp"), size=4096, cached=False)
            xicsics_region = soc_region_cls(origin=soc.mem_map.get("xicsics"), size=4096, cached=False)
            soc.bus.add_slave(name="xicsicp", slave=self.xics.icp_bus, region=xicsicp_region)
            soc.bus.add_slave(name="xicsics", slave=self.xics.ics_bus, region=xicsics_region)

    @staticmethod
    def add_sources(platform, use_ghdl_yosys_plugin=False):
        sources = [
            # Common / Types / Helpers.
            "decode_types.vhdl",
            "wishbone_types.vhdl",
            "utils.vhdl",
            "common.vhdl",
            "helpers.vhdl",

            # Fetch.
            "fetch1.vhdl",

            # Instruction/Data Cache.
            "cache_ram.vhdl",
            "plru.vhdl",
            "dcache.vhdl",
            "icache.vhdl",

            # Decode.
            "insn_helpers.vhdl",
            "decode1.vhdl",
            "gpr_hazard.vhdl",
            "cr_hazard.vhdl",
            "control.vhdl",
            "decode2.vhdl",

            # Register/CR File.
            "register_file.vhdl",
            "crhelpers.vhdl",
            "cr_file.vhdl",

            # Execute.
            "ppc_fx_insns.vhdl",
            "logical.vhdl",
            "rotator.vhdl",
            "countzero.vhdl",
            "execute1.vhdl",

            # Load/Store.
            "loadstore1.vhdl",

            # Multiply/Divide.
            "multiply.vhdl",
            "divider.vhdl",

            # Writeback.
            "writeback.vhdl",

            # MMU.
            "mmu.vhdl",

            # Core.
            "core_debug.vhdl",
            "core.vhdl",
        ]
        sdir = get_data_mod("cpu", "microwatt").data_location
        cdir = os.path.dirname(__file__)
        # Convert VHDL to Verilog through GHDL/Yosys.
        if use_ghdl_yosys_plugin:
            from litex.build import tools
            import subprocess
            ys = []
            ys.append("ghdl --ieee=synopsys -fexplicit -frelaxed-rules --std=08 \\")
            for source in sources:
                ys.append(os.path.join(sdir, source) + " \\")
            ys.append(os.path.join(os.path.dirname(__file__), "microwatt_wrapper.vhdl") + " \\")
            ys.append("-e microwatt_wrapper")
            ys.append("chformal -assert -remove")
            ys.append("write_verilog {}".format(os.path.join(cdir, "microwatt.v")))
            tools.write_to_file(os.path.join(cdir, "microwatt.ys"), "\n".join(ys))
            if subprocess.call(["yosys", "-q", "-m", "ghdl", os.path.join(cdir, "microwatt.ys")]):
                raise OSError("Unable to convert Microwatt CPU to verilog, please check your GHDL-Yosys-plugin install")
            platform.add_source(os.path.join(cdir, "microwatt.v"))
        # Direct use of VHDL sources.
        else:
            platform.add_sources(sdir, *sources)
            platform.add_source(os.path.join(os.path.dirname(__file__), "microwatt_wrapper.vhdl"))

    def do_finalize(self):
        self.specials += Instance("microwatt_wrapper", **self.cpu_params)

# XICS Slave ---------------------------------------------------------------------------------------

class XICSSlave(Module, AutoCSR):
    def __init__(self, platform, core_irq_out=Signal(), int_level_in=Signal(16), variant="standard"):
        self.variant = variant

        self.icp_bus = icp_bus = wishbone.Interface(data_width=32, adr_width=12)
        self.ics_bus = ics_bus = wishbone.Interface(data_width=32, adr_width=12)

        # XICS signals.
        self.ics_icp_xfer_src = Signal(4)
        self.ics_icp_xfer_pri = Signal(8)

        self.icp_params = dict(
            # Clk / Rst.
            i_clk            = ClockSignal("sys"),
            i_rst            = ResetSignal("sys"),

            # Wishbone Bus.
            o_wishbone_dat_r = icp_bus.dat_r,
            o_wishbone_ack   = icp_bus.ack,

            i_wishbone_adr   = icp_bus.adr,
            i_wishbone_dat_w = icp_bus.dat_w,
            i_wishbone_cyc   = icp_bus.cyc,
            i_wishbone_stb   = icp_bus.stb,
            i_wishbone_sel   = icp_bus.sel,
            i_wishbone_we    = icp_bus.we,

            i_ics_in_src     = self.ics_icp_xfer_src,
            i_ics_in_pri     = self.ics_icp_xfer_pri,

            o_core_irq_out   = core_irq_out,
        )

        self.ics_params = dict(
            # Clk / Rst.
            i_clk            = ClockSignal("sys"),
            i_rst            = ResetSignal("sys"),

            # Wishbone Bus.
            o_wishbone_dat_r = ics_bus.dat_r,
            o_wishbone_ack   = ics_bus.ack,

            i_wishbone_adr   = ics_bus.adr,
            i_wishbone_dat_w = ics_bus.dat_w,
            i_wishbone_cyc   = ics_bus.cyc,
            i_wishbone_stb   = ics_bus.stb,
            i_wishbone_sel   = ics_bus.sel,
            i_wishbone_we    = ics_bus.we,

            i_int_level_in   = int_level_in,

            o_icp_out_src    = self.ics_icp_xfer_src,
            o_icp_out_pri    = self.ics_icp_xfer_pri,
        )

        # Add VHDL sources.
        self.add_sources(platform, use_ghdl_yosys_plugin="ghdl" in self.variant)

    @staticmethod
    def add_sources(platform, use_ghdl_yosys_plugin=False):
        sources = [
            # Common / Types / Helpers
            "decode_types.vhdl",
            "wishbone_types.vhdl",
            "utils.vhdl",
            "common.vhdl",
            "helpers.vhdl",

            # XICS controller
            "xics.vhdl",
        ]
        sdir = get_data_mod("cpu", "microwatt").data_location
        cdir = os.path.dirname(__file__)
        if use_ghdl_yosys_plugin:
            from litex.build import tools
            import subprocess

            # ICP
            ys = []
            ys.append("ghdl --ieee=synopsys -fexplicit -frelaxed-rules --std=08 \\")
            for source in sources:
                ys.append(os.path.join(sdir, source) + " \\")
            ys.append(os.path.join(os.path.dirname(__file__), "xics_wrapper.vhdl") + " \\")
            ys.append("-e xics_icp_wrapper")
            ys.append("chformal -assert -remove")
            ys.append("write_verilog {}".format(os.path.join(cdir, "xics_icp.v")))
            tools.write_to_file(os.path.join(cdir, "xics_icp.ys"), "\n".join(ys))
            if subprocess.call(["yosys", "-q", "-m", "ghdl", os.path.join(cdir, "xics_icp.ys")]):
                raise OSError("Unable to convert Microwatt XICS ICP controller to verilog, please check your GHDL-Yosys-plugin install")
            platform.add_source(os.path.join(cdir, "xics_icp.v"))

            # ICS
            ys = []
            ys.append("ghdl --ieee=synopsys -fexplicit -frelaxed-rules --std=08 \\")
            for source in sources:
                ys.append(os.path.join(sdir, source) + " \\")
            ys.append(os.path.join(os.path.dirname(__file__), "xics_wrapper.vhdl") + " \\")
            ys.append("-e xics_ics_wrapper")
            ys.append("chformal -assert -remove")
            ys.append("write_verilog {}".format(os.path.join(cdir, "xics_ics.v")))
            tools.write_to_file(os.path.join(cdir, "xics_ics.ys"), "\n".join(ys))
            if subprocess.call(["yosys", "-q", "-m", "ghdl", os.path.join(cdir, "xics_ics.ys")]):
                raise OSError("Unable to convert Microwatt XICS ICP controller to verilog, please check your GHDL-Yosys-plugin install")
            platform.add_source(os.path.join(cdir, "xics_ics.v"))
        else:
            platform.add_sources(sdir, *sources)
            platform.add_source(os.path.join(os.path.dirname(__file__), "xics_wrapper.vhdl"))

    def do_finalize(self):
        self.specials += Instance("xics_icp_wrapper", **self.icp_params)
        self.specials += Instance("xics_ics_wrapper", **self.ics_params)
